#!/usr/bin/env python
"""
The tools.visualisation module contains functions for aiding visualisation of data.
"""

# Import modules
import os
import shutil
import subprocess
from typing import Union, List

import rsgislib
import rsgislib.imageutils
import rsgislib.tools.utils
import rsgislib.tools.filetools
import rsgislib.tools.projection
import rsgislib.tools.tilecacheutils


def create_kmz_img(
    input_img: str,
    output_file: str,
    bands: str,
    reproj_wgs84: bool = True,
    finite_msk: bool = False,
    tmp_dir: str = None,
):
    """
    A function to convert an input image to a KML/KMZ file, where the input image
    is stretched and bands sub-selected / ordered as required for visualisation.

    :param input_img: input image file (any format that gdal supports)
    :param output_file: output image file (extension kmz for KMZ output /
                        kml for KML output)
    :param bands: a string (comma seperated) with the bands to be selected.
                  (e.g., '1', '1,2,3', '5,6,4')
    :param reproj_wgs84: specify whether the image should be explicitly reprojected
                         to WGS84 Lat/Long before transformation to KML.
    :param finite_msk: specify whether the image data should be masked so all
                       values are finite before stretching.
    :param tmp_dir: A temp directory path for processing files.

    """
    if not rsgislib.tools.filetools.does_path_exists_or_creatable(output_file):
        raise rsgislib.RSGISPyException(
            f"Output file path is not creatable: {output_file}"
        )

    band_lst = bands.split(",")
    if len(band_lst) == 3:
        multi_band = True
    elif len(band_lst) == 1:
        multi_band = False
    else:
        raise rsgislib.RSGISPyException(
            "You need to either provide 1 or 3 bands: {}".format(band_lst)
        )

    n_img_bands = rsgislib.imageutils.get_img_band_count(input_img)

    uid_str = rsgislib.tools.utils.uid_generator()
    if tmp_dir is None:
        tmp_dir = os.path.join(os.path.dirname(input_img), uid_str)
    else:
        tmp_dir = os.path.join(tmp_dir, uid_str)
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
    base_name = rsgislib.tools.filetools.get_file_basename(input_img)

    sel_img_bands_img = ""
    if (n_img_bands == 1) and (not multi_band):
        sel_img_bands_img = input_img
    elif (
        (n_img_bands == 3)
        and (multi_band)
        and (band_lst[0] == "1")
        and (band_lst[1] == "2")
        and (band_lst[2] == "3")
    ):
        sel_img_bands_img = input_img
    else:
        s_bands = []
        for str_band in band_lst:
            s_bands.append(int(str_band))
        sel_img_bands_img = os.path.join(tmp_dir, f"{base_name}_sband.kea")
        rsgislib.imageutils.select_img_bands(
            input_img,
            sel_img_bands_img,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            s_bands,
        )

    img_to_strch = sel_img_bands_img
    if finite_msk:
        finite_msk_img = os.path.join(tmp_dir, f"{base_name}_FiniteMsk.kea")
        rsgislib.imageutils.gen_finite_mask(sel_img_bands_img, finite_msk_img, "KEA")
        img_to_strch = os.path.join(tmp_dir, f"{base_name}_Msk2FiniteRegions.kea")
        rsgislib.imageutils.mask_img(
            sel_img_bands_img,
            finite_msk_img,
            img_to_strch,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            0,
            0,
        )

    stretch_img = os.path.join(tmp_dir, f"{base_name}_stretch.kea")
    rsgislib.imageutils.stretch_img(
        img_to_strch,
        stretch_img,
        False,
        "",
        True,
        False,
        "KEA",
        rsgislib.TYPE_8UINT,
        rsgislib.imageutils.STRETCH_LINEARSTDDEV,
        2,
    )

    gdal_input_img = stretch_img
    if reproj_wgs84:
        lat_long_img = os.path.join(tmp_dir, f"{base_name}_latlong.kea")
        out_wkt = os.path.join(tmp_dir, f"{base_name}_latlong.wkt")
        wgs84_wkt = rsgislib.tools.projection.get_wkt_from_epsg_code(4326)
        rsgislib.tools.utils.write_list_to_file([wgs84_wkt], out_wkt)
        rsgislib.imageutils.reproject_image(
            stretch_img,
            lat_long_img,
            out_wkt,
            gdalformat="KEA",
            interp_method=rsgislib.INTERP_CUBIC,
            in_wkt=None,
            no_data_val=0.0,
            out_pxl_res="auto",
            snap_to_grid=True,
        )
        gdal_input_img = lat_long_img

    cmd = ["gdal_translate", "-of", "KMLSUPEROVERLAY", gdal_input_img, output_file]
    print(cmd)
    try:
        subprocess.run(cmd, check=True)
    except OSError as e:
        raise rsgislib.RSGISPyException("Could not execute command: {}".format(cmd))

    shutil.rmtree(tmp_dir)


def create_webtiles_img_no_stats_msk(
    input_img: str,
    out_dir: str,
    bands: str,
    zoom_levels: str = "2-10",
    resample="average",
    finite_msk: bool = False,
    tms: bool = True,
    tmp_dir=None,
):
    """
    A function to convert an input image to a tile cache for web map servers, where the input image
    is stretched and bands sub-selected / ordered as required for visualisation.

    :param input_img: input image file (any format that gdal supports)
    :param out_dir: output directory within which the cache will be created.
    :param bands: a string (comma seperated) with the bands to be selected. (e.g., '1', '1,2,3', '5,6,4')
    :param zoom_levels: The zoom levels to be created for the web tile cache.
    :param resample: Method of resampling (average,near,bilinear,cubic,cubicspline,lanczos,antialias)
    :param finite_msk: specify whether the image data should be masked so all values are finite before stretching.
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
                (TMS: gdal2tiles.py native. XYZ: GIS Compatible)
    :param tmp_dir: A temp directory path for processing files.

    """
    if not rsgislib.tools.filetools.does_path_exists_or_creatable(out_dir):
        raise rsgislib.RSGISPyException(f"Output file path is not creatable: {out_dir}")

    band_lst = bands.split(",")
    multi_band = False
    if len(band_lst) == 3:
        multi_band = True
    elif len(band_lst) == 1:
        multi_band = False
    else:
        raise rsgislib.RSGISPyException(
            "You need to either provide 1 or 3 bands: {}".format(band_lst)
        )
    n_img_bands = rsgislib.imageutils.get_img_band_count(input_img)

    uid_str = rsgislib.tools.utils.uid_generator()

    if tmp_dir is None:
        tmp_dir = os.path.join(os.path.dirname(input_img), uid_str)
    else:
        tmp_dir = os.path.join(tmp_dir, uid_str)
    os.makedirs(tmp_dir)
    base_name = rsgislib.tools.filetools.get_file_basename(input_img)

    sel_img_bands_img = ""
    if (n_img_bands == 1) and (not multi_band):
        sel_img_bands_img = input_img
    elif (
        (n_img_bands == 3)
        and (multi_band)
        and (band_lst[0] == "1")
        and (band_lst[1] == "2")
        and (band_lst[2] == "3")
    ):
        sel_img_bands_img = input_img
    else:
        s_bands = []
        for str_band in band_lst:
            s_bands.append(int(str_band))
        sel_img_bands_img = os.path.join(tmp_dir, f"{base_name}_sband.kea")
        rsgislib.imageutils.select_img_bands(
            input_img,
            sel_img_bands_img,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            s_bands,
        )

    img_to_strch = sel_img_bands_img
    if finite_msk:
        finite_msk_img = os.path.join(tmp_dir, f"{base_name}_FiniteMsk.kea")
        rsgislib.imageutils.gen_finite_mask(sel_img_bands_img, finite_msk_img, "KEA")
        img_to_strch = os.path.join(tmp_dir, f"{base_name}_Msk2FiniteRegions.kea")
        rsgislib.imageutils.mask_img(
            sel_img_bands_img,
            finite_msk_img,
            img_to_strch,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            0,
            0,
        )

    stretch_img = os.path.join(tmp_dir, f"{base_name}_stretch.kea")
    rsgislib.imageutils.stretch_img(
        img_to_strch,
        stretch_img,
        False,
        "",
        True,
        False,
        "KEA",
        rsgislib.TYPE_8UINT,
        rsgislib.imageutils.STRETCH_LINEARSTDDEV,
        2,
    )

    cmd = [
        "gdal2tiles.py",
        "-r",
        resample,
        "-z",
        zoom_levels,
        "-a",
        "0",
        stretch_img,
        out_dir,
    ]
    print(cmd)
    try:
        subprocess.run(cmd, check=True)
    except OSError as e:
        raise rsgislib.RSGISPyException("Could not execute command: {}".format(cmd))

    if not tms:
        rsgislib.tools.tilecacheutils.convert_between_tms_xyz(out_dir)

    shutil.rmtree(tmp_dir)


def create_webtiles_img(
    input_img: str,
    bands: str,
    out_dir: str,
    zoom_levels: str = "2-10",
    img_stats_msk: str = None,
    img_msk_vals: int = 1,
    tmp_dir: str = None,
    webview: bool = True,
    tms: bool = True,
    no_data_val: float = None,
):
    """
    A function to produce a web cache for the input image.

    :param input_img: input image file (any format that gdal supports)
    :param bands: a string (comma seperated) with the bands to be selected.
                  (e.g., '1', '1,2,3', '5,6,4')
    :param out_dir: output directory within which the cache will be created.
    :param zoom_levels: The zoom levels to be created for the web tile cache.
    :param img_stats_msk: Optional (default=None) input image which is used to define
                          regions calculate the image stats for stretch.
    :param img_msk_vals: The pixel(s) value define the region of interest in the
                         image mask (can be list of values or single value).
    :param tmp_dir: an input directory which can be used to write temporary files /
                    directories. If not provided (i.e., input is None) then a local
                    directory will be define in the same folder as the input image.
    :param webview: Provide default GDAL leaflet web viewer.
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
                (TMS: gdal2tiles.py native. XYZ: GIS Compatible)
    :param no_data_val: Optionally provide an input image no data value otherwise
                        it will be read from the input image header.

    """
    if not rsgislib.tools.filetools.does_path_exists_or_creatable(out_dir):
        raise rsgislib.RSGISPyException(f"Output file path is not creatable: {out_dir}")

    band_lst = bands.split(",")
    multi_band = False
    if len(band_lst) == 3:
        multi_band = True
    elif len(band_lst) == 1:
        multi_band = False
    else:
        raise rsgislib.RSGISPyException(
            "You need to either provide 1 or 3 bands. {}".format(band_lst)
        )
    n_img_bands = rsgislib.imageutils.get_img_band_count(input_img)
    if no_data_val is None:
        no_data_val = rsgislib.imageutils.get_img_no_data_value(input_img)
    if no_data_val is None:
        raise rsgislib.RSGISPyException("No data value is not defined.")

    uid_str = rsgislib.tools.utils.uid_generator()
    if tmp_dir is not None:
        tmp_dir = os.path.join(tmp_dir, uid_str)
    else:
        tmp_dir = os.path.join(os.path.dirname(input_img), uid_str)
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
    base_name = rsgislib.tools.filetools.get_file_basename(input_img)

    sel_img_bands_img = ""
    if (n_img_bands == 1) and (not multi_band):
        sel_img_bands_img = input_img
    elif (
        (n_img_bands == 3)
        and (multi_band)
        and (band_lst[0] == "1")
        and (band_lst[1] == "2")
        and (band_lst[2] == "3")
    ):
        sel_img_bands_img = input_img
    else:
        s_bands = []
        for str_band in band_lst:
            s_bands.append(int(str_band))
        sel_img_bands_img = os.path.join(tmp_dir, f"{base_name}_sband.kea")
        rsgislib.imageutils.select_img_bands(
            input_img,
            sel_img_bands_img,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            s_bands,
        )

    img_to_strch = sel_img_bands_img
    stretch_img = os.path.join(tmp_dir, f"{base_name}_stretch.kea")
    if img_stats_msk is not None:
        img_to_strch_mskd = os.path.join(tmp_dir, f"{base_name}_MskdImg.kea")
        rsgislib.imageutils.mask_img(
            sel_img_bands_img,
            img_stats_msk,
            img_to_strch_mskd,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            no_data_val,
            img_msk_vals,
        )
        stretch_img_stats = os.path.join(tmp_dir, f"{base_name}_stretch_statstmp.txt")
        stretch_img_tmp = os.path.join(tmp_dir, f"{base_name}_stretch_tmp.kea")
        rsgislib.imageutils.stretch_img(
            img_to_strch_mskd,
            stretch_img_tmp,
            True,
            stretch_img_stats,
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

        rsgislib.imageutils.stretch_img_with_stats(
            img_to_strch,
            stretch_img,
            stretch_img_stats,
            "KEA",
            rsgislib.TYPE_8UINT,
            no_data_val,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )
    else:
        rsgislib.imageutils.stretch_img(
            img_to_strch,
            stretch_img,
            False,
            "",
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

    webview_opt = "none"
    if webview:
        webview_opt = "leaflet"

    cmd = [
        "gdal2tiles.py",
        "-r",
        "average",
        "-z",
        zoom_levels,
        "-a",
        "0",
        "-w",
        webview_opt,
        stretch_img,
        out_dir,
    ]
    print(cmd)
    try:
        subprocess.run(cmd, check=True)
    except OSError as e:
        raise rsgislib.RSGISPyException("Could not execute command: {}".format(cmd))
    shutil.rmtree(tmp_dir)

    if not tms:
        rsgislib.tools.tilecacheutils.convert_between_tms_xyz(out_dir)


def create_quicklook_imgs(
    input_img: str,
    bands: str,
    output_imgs: Union[str, List[str]] = "quicklook.jpg",
    output_img_sizes: Union[int, List[int]] = 250,
    scale_axis: str = "auto",
    img_stats_msk: str = None,
    img_msk_vals: int = 1,
    stretch_file: str = None,
    tmp_dir: str = None,
    no_data_val: float = None,
):
    """
    A function to produce a quicklook image(s).

    :param input_img: input image file (any format that gdal supports)
    :param bands: a string (comma seperated) with the bands to be selected.
                  (e.g., '1', '1,2,3', '5,6,4')
    :param output_imgs: a single output image or list of output images. The same
                        size as output_img_sizes.
    :param output_img_sizes: the output image size (in pixels) or list of output
                             image sizes.
    :param scale_axis: the axis to which the output_img_sizes refer. Options:
                       width, height or auto. Auto applies the output_img_sizes to
                       the longest of the two axes.
    :param img_stats_msk: Optional (default=None) input image which is used to
                          define regions calculate the image stats for stretch.
    :param img_msk_vals: The pixel(s) value define the region of interest in the
                         image mask (can be list of values or single value).
    :param stretch_file: a stretch stats file to standardise the stretch between
                         a number of input files.
    :param tmp_dir: an input directory which can be used to write temporary files /
                    directories. If not provided (i.e., input is None) then a local
                    directory will be define in the same folder as the input image.
    :param no_data_val: Optionally provide an input image no data value otherwise
                        it will be read from the input image header.

    """
    import rsgislib.tools.utils

    if scale_axis not in ["width", "height", "auto"]:
        raise rsgislib.RSGISPyException(
            "Input parameter 'scale_axis' must have the "
            "value 'width', 'height' or 'auto'."
        )

    n_out_imgs = 1
    if type(output_imgs) is list:
        n_out_imgs = len(output_imgs)
        if type(output_img_sizes) is not list:
            raise rsgislib.RSGISPyException(
                "If the output_imgs input is a list so must output_img_sizes."
            )
        if len(output_img_sizes) != n_out_imgs:
            raise rsgislib.RSGISPyException(
                "output_imgs and output_img_sizes must be the same length"
            )

        if n_out_imgs == 1:
            output_imgs = output_imgs[0]
            output_img_sizes = output_img_sizes[0]

    band_lst = bands.split(",")
    multi_band = False
    if len(band_lst) == 3:
        multi_band = True
    elif len(band_lst) == 1:
        multi_band = False
    else:
        raise rsgislib.RSGISPyException(
            "You need to either provide 1 or 3 bands: {}".format(band_lst)
        )
    n_img_bands = rsgislib.imageutils.get_img_band_count(input_img)
    if no_data_val is None:
        no_data_val = rsgislib.imageutils.get_img_no_data_value(input_img)
    if no_data_val is None:
        raise rsgislib.RSGISPyException("No data value is not defined.")

    uid_str = rsgislib.tools.utils.uid_generator()
    if tmp_dir is not None:
        tmp_dir = os.path.join(tmp_dir, uid_str)
    else:
        tmp_dir = os.path.join(os.path.dirname(input_img), uid_str)
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
    base_name = rsgislib.tools.filetools.get_file_basename(input_img)

    sel_img_bands_img = ""
    if (n_img_bands == 1) and (not multi_band):
        sel_img_bands_img = input_img
    elif (
        (n_img_bands == 3)
        and (multi_band)
        and (band_lst[0] == "1")
        and (band_lst[1] == "2")
        and (band_lst[2] == "3")
    ):
        sel_img_bands_img = input_img
    else:
        s_bands = []
        for str_band in band_lst:
            s_bands.append(int(str_band))
        sel_img_bands_img = os.path.join(tmp_dir, f"{base_name}_sband.kea")
        rsgislib.imageutils.select_img_bands(
            input_img,
            sel_img_bands_img,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            s_bands,
        )

    img_to_strch = sel_img_bands_img
    stretch_img = os.path.join(tmp_dir, f"{base_name}_stretch.kea")
    if stretch_file is not None:
        rsgislib.imageutils.stretch_img_with_stats(
            img_to_strch,
            stretch_img,
            stretch_file,
            "KEA",
            rsgislib.TYPE_8UINT,
            no_data_val,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )
    elif img_stats_msk is not None:
        img_to_strch_mskd = os.path.join(tmp_dir, f"{base_name}_MskdImg.kea")
        rsgislib.imageutils.mask_img(
            sel_img_bands_img,
            img_stats_msk,
            img_to_strch_mskd,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            no_data_val,
            img_msk_vals,
        )
        stretch_img_stats = os.path.join(tmp_dir, f"{base_name}_stretch_statstmp.txt")
        stretch_img_tmp = os.path.join(tmp_dir, f"{base_name}_stretch_tmp.kea")
        rsgislib.imageutils.stretch_img(
            img_to_strch_mskd,
            stretch_img_tmp,
            True,
            stretch_img_stats,
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

        rsgislib.imageutils.stretch_img_with_stats(
            img_to_strch,
            stretch_img,
            stretch_img_stats,
            "KEA",
            rsgislib.TYPE_8UINT,
            no_data_val,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )
    else:
        rsgislib.imageutils.stretch_img(
            img_to_strch,
            stretch_img,
            False,
            "",
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

    if scale_axis == "auto":
        x_size, y_size = rsgislib.imageutils.get_img_size(stretch_img)
        if x_size > y_size:
            scale_axis = "width"
        else:
            scale_axis = "height"

    if n_out_imgs == 1:
        if not rsgislib.tools.filetools.does_path_exists_or_creatable(output_imgs):
            raise rsgislib.RSGISPyException(
                f"Output file path is not creatable: {output_imgs}"
            )
        cmd = ["gdal_translate", "-of", "JPEG", "-ot", "Byte"]
        if scale_axis == "width":
            cmd.append("-outsize")
            cmd.append(f"{int(output_img_sizes)}")
            cmd.append("0")
        else:
            cmd.append("-outsize")
            cmd.append("0")
            cmd.append(f"{int(output_img_sizes)}")
        cmd.append("-r")
        cmd.append("average")
        cmd.append(stretch_img)
        cmd.append(output_imgs)

        print(cmd)
        try:
            subprocess.run(cmd, check=True)
        except OSError as e:
            raise rsgislib.RSGISPyException("Could not execute command: {}".format(cmd))

    elif n_out_imgs > 1:
        for i in range(n_out_imgs):
            if not rsgislib.tools.filetools.does_path_exists_or_creatable(
                output_imgs[i]
            ):
                raise rsgislib.RSGISPyException(
                    f"Output file path is not creatable: {output_imgs}"
                )
            cmd = ["gdal_translate", "-of", "JPEG", "-ot", "Byte"]
            if scale_axis == "width":
                cmd.append("-outsize")
                cmd.append(f"{int(output_img_sizes[i])}")
                cmd.append("0")
            else:
                cmd.append("-outsize")
                cmd.append("0")
                cmd.append(f"{int(output_img_sizes[i])}")
            cmd.append("-r")
            cmd.append("average")
            cmd.append(stretch_img)
            cmd.append(output_imgs[i])

            print(cmd)
            try:
                subprocess.run(cmd, check=True)
            except OSError as e:
                raise rsgislib.RSGISPyException(
                    "Could not execute command: {}".format(cmd)
                )
    shutil.rmtree(tmp_dir)


def create_mbtile_file(
    input_img: str,
    bands: str,
    output_mbtiles: str,
    scale_input_img: int = 50,
    img_stats_msk: str = None,
    img_msk_vals: int = 1,
    tmp_dir: str = None,
    tile_format: str = "PNG",
    no_data_val: float = None,
):
    """
    A function to produce an MBTile file for visualisation.

    :param input_img: input image file (any format that gdal supports)
    :param bands: a string (comma seperated) with the bands to be selected.
                  (e.g., '1', '1,2,3', '5,6,4')
    :param output_mbtiles: output MBTiles file which will be created.
    :param scale_input_img: The scale of the output image with respect to the input
                            as a percentage (e.g., 50% reduction in size).
    :param img_stats_msk: Optional (default=None) input image which is used to
                          define regions calculate the image stats for stretch.
    :param img_msk_vals: The pixel(s) value define the region of interest in the
                         image mask (can be list of values or single value).
    :param tmp_dir: an input directory which can be used to write temporary files /
                    directories. If not provided (i.e., input is None) then a local
                    directory will be define in the same folder as the input image.
    :param tile_format: Specify the tile file format to use, options: PNG, PNG8
                        and JPG. Default: PNG
    :param no_data_val: Optionally provide an input image no data value otherwise
                        it will be read from the input image header.

    """
    from osgeo import gdal

    band_lst = bands.split(",")
    multi_band = False
    if len(band_lst) == 3:
        multi_band = True
    elif len(band_lst) == 1:
        multi_band = False
    else:
        raise rsgislib.RSGISPyException(
            "You need to either provide 1 or 3 bands: {}".format(band_lst)
        )
    n_img_bands = rsgislib.imageutils.get_img_band_count(input_img)
    if no_data_val is None:
        no_data_val = rsgislib.imageutils.get_img_no_data_value(input_img)
    if no_data_val is None:
        raise rsgislib.RSGISPyException("No data value is not defined.")

    uid_str = rsgislib.tools.utils.uid_generator()
    if tmp_dir is not None:
        tmp_dir = os.path.join(tmp_dir, uid_str)
    else:
        tmp_dir = os.path.join(os.path.dirname(input_img), uid_str)
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
    base_name = rsgislib.tools.filetools.get_file_basename(input_img)

    sel_img_bands_img = ""
    if (n_img_bands == 1) and (not multi_band):
        sel_img_bands_img = input_img
    elif (
        (n_img_bands == 3)
        and (multi_band)
        and (band_lst[0] == "1")
        and (band_lst[1] == "2")
        and (band_lst[2] == "3")
    ):
        sel_img_bands_img = input_img
    else:
        s_bands = []
        for str_band in band_lst:
            s_bands.append(int(str_band))
        sel_img_bands_img = os.path.join(tmp_dir, f"{base_name}_sband.kea")
        rsgislib.imageutils.select_img_bands(
            input_img,
            sel_img_bands_img,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            s_bands,
        )

    img_to_strch = sel_img_bands_img
    stretch_img = os.path.join(tmp_dir, f"{base_name}_stretch.kea")
    if img_stats_msk is not None:
        img_to_strch_mskd = os.path.join(tmp_dir, f"{base_name}_MskdImg.kea")
        rsgislib.imageutils.mask_img(
            sel_img_bands_img,
            img_stats_msk,
            img_to_strch_mskd,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            no_data_val,
            img_msk_vals,
        )
        stretch_img_stats = os.path.join(tmp_dir, f"{base_name}_stretch_statstmp.txt")
        stretch_img_tmp = os.path.join(tmp_dir, f"{base_name}_stretch_tmp.kea")
        rsgislib.imageutils.stretch_img(
            img_to_strch_mskd,
            stretch_img_tmp,
            True,
            stretch_img_stats,
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )
        rsgislib.imageutils.stretch_img_with_stats(
            img_to_strch,
            stretch_img,
            stretch_img_stats,
            "KEA",
            rsgislib.TYPE_8UINT,
            no_data_val,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )
    else:
        rsgislib.imageutils.stretch_img(
            img_to_strch,
            stretch_img,
            False,
            "",
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

    stretch_img_re_proj = os.path.join(tmp_dir, f"{base_name}_stretch_epsg3857.kea")
    rsgislib.imageutils.gdal_warp(
        stretch_img,
        stretch_img_re_proj,
        3857,
        interp_method=rsgislib.INTERP_NEAREST_NEIGHBOUR,
        gdalformat="KEA",
        options=[],
    )

    try:
        import tqdm

        pbar = tqdm.tqdm(total=100)
        callback = lambda *args, **kw: pbar.update()
    except:
        callback = gdal.TermProgress

    trans_opt = gdal.TranslateOptions(
        format="MBTiles",
        widthPct=scale_input_img,
        heightPct=scale_input_img,
        noData=0,
        options=["TILE_FORMAT={}".format(tile_format)],
        callback=callback,
    )
    gdal.Translate(output_mbtiles, stretch_img_re_proj, options=trans_opt)
    rsgislib.imageutils.pop_img_stats(
        output_mbtiles, use_no_data=True, no_data_val=0, calc_pyramids=True
    )
    shutil.rmtree(tmp_dir)


def create_webtiles_vis_gtiff_img(
    input_img: str,
    bands: str,
    output_dir: str,
    scaled_gtiff_img: str,
    zoom_levels: str = "2-10",
    img_stats_msk: str = None,
    img_msk_vals: int = 1,
    stretch_file: str = None,
    tmp_dir: str = None,
    webview: bool = True,
    scale: float = 0,
    tms: bool = True,
    no_data_val: float = None,
):
    """
    A function to produce web cache and scaled and stretched geotiff.

    :param input_img: input image file (any format that gdal supports)
    :param bands: a string (comma seperated) with the bands to be selected.
                  (e.g., '1', '1,2,3', '5,6,4')
    :param output_dir: output directory within which the cache will be created.
    :param scaled_gtiff_img: output geotiff image path and file name.
    :param zoom_levels: The zoom levels to be created for the web tile cache.
    :param img_stats_msk: Optional (default=None) input image which is used to
                          define regions calculate the image stats for stretch.
    :param img_msk_vals: The pixel(s) value define the region of interest in the
                         image mask (can be list of values or single value).
    :param stretch_file: a stretch stats file to standardise the stretch between
                         a number of input files.
    :param tmp_dir: an input directory which can be used to write temporary files /
                    directories. If not provided (i.e., input is None) then a local
                    directory will be define in the same folder as the input image.
    :param webview: Provide default GDAL leaflet web viewer.
    :param scale: the scale output geotiff. Input is percentage in the x-axis.
                  If zero (default) then no scaling will be applied.
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
                (TMS: gdal2tiles.py native. XYZ: GIS Compatible)
    :param no_data_val: Optionally provide an input image no data value otherwise
                        it will be read from the input image header.

    """
    if not rsgislib.tools.filetools.does_path_exists_or_creatable(output_dir):
        raise rsgislib.RSGISPyException(
            f"Output file path is not creatable: {output_dir}"
        )
    if not rsgislib.tools.filetools.does_path_exists_or_creatable(scaled_gtiff_img):
        raise rsgislib.RSGISPyException(
            f"Output file path is not creatable: {scaled_gtiff_img}"
        )

    band_lst = bands.split(",")
    multi_band = False
    if len(band_lst) == 3:
        multi_band = True
    elif len(band_lst) == 1:
        multi_band = False
    else:
        raise rsgislib.RSGISPyException(
            "You need to either provide 1 or 3 bands: {}".format(band_lst)
        )
    n_img_bands = rsgislib.imageutils.get_img_band_count(input_img)
    if no_data_val is None:
        no_data_val = rsgislib.imageutils.get_img_no_data_value(input_img)
    if no_data_val is None:
        raise rsgislib.RSGISPyException("No data value is not defined.")

    uid_str = rsgislib.tools.utils.uid_generator()
    if tmp_dir is not None:
        tmp_dir = os.path.join(tmp_dir, uid_str)
    else:
        tmp_dir = os.path.join(os.path.dirname(input_img), uid_str)
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
    base_name = rsgislib.tools.filetools.get_file_basename(input_img)

    sel_img_bands_img = ""
    if (n_img_bands == 1) and (not multi_band):
        sel_img_bands_img = input_img
    elif (
        (n_img_bands == 3)
        and (multi_band)
        and (band_lst[0] == "1")
        and (band_lst[1] == "2")
        and (band_lst[2] == "3")
    ):
        sel_img_bands_img = input_img
    else:
        s_bands = []
        for str_band in band_lst:
            s_bands.append(int(str_band))
        sel_img_bands_img = os.path.join(tmp_dir, f"{base_name}_sband.kea")
        rsgislib.imageutils.select_img_bands(
            input_img,
            sel_img_bands_img,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            s_bands,
        )

    img_to_strch = sel_img_bands_img
    stretch_img = os.path.join(tmp_dir, f"{base_name}_stretch.kea")
    if stretch_file is not None:
        rsgislib.imageutils.stretch_img_with_stats(
            img_to_strch,
            stretch_img,
            stretch_file,
            "KEA",
            rsgislib.TYPE_8UINT,
            no_data_val,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )
    elif img_stats_msk is not None:
        img_to_strch_mskd = os.path.join(tmp_dir, f"{base_name}_MskdImg.kea")
        rsgislib.imageutils.mask_img(
            sel_img_bands_img,
            img_stats_msk,
            img_to_strch_mskd,
            "KEA",
            rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img),
            no_data_val,
            img_msk_vals,
        )
        stretch_img_stats = os.path.join(tmp_dir, f"{base_name}_stretch_statstmp.txt")
        stretch_img_tmp = os.path.join(tmp_dir, f"{base_name}_stretch_tmp.kea")
        rsgislib.imageutils.stretch_img(
            img_to_strch_mskd,
            stretch_img_tmp,
            True,
            stretch_img_stats,
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

        rsgislib.imageutils.stretch_img_with_stats(
            img_to_strch,
            stretch_img,
            stretch_img_stats,
            "KEA",
            rsgislib.TYPE_8UINT,
            no_data_val,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )
    else:
        rsgislib.imageutils.stretch_img(
            img_to_strch,
            stretch_img,
            False,
            "",
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

    webview_opt = "none"
    if webview:
        webview_opt = "leaflet"

    cmd = [
        "gdal2tiles.py",
        "-r",
        "average",
        "-z",
        zoom_levels,
        "-a",
        "0",
        "-w",
        webview_opt,
        stretch_img,
        output_dir,
    ]
    print(cmd)
    try:
        subprocess.run(cmd, check=True)
    except OSError as e:
        raise rsgislib.RSGISPyException("Could not execute command: {}".format(cmd))

    if not tms:
        rsgislib.tools.tilecacheutils.convert_between_tms_xyz(output_dir)

    if scale > 0:
        cmd = [
            "gdal_translate",
            "-of",
            "GTIFF",
            "-co",
            "TILED=YES",
            "-co",
            "COMPRESS=JPEG",
            "-co",
            "BIGTIFF=NO",
            "-ot",
            "Byte",
            "-outsize",
            f"{scale}%",
            "0",
            "-r",
            "average",
            stretch_img,
            scaled_gtiff_img,
        ]
    else:
        cmd = [
            "gdal_translate",
            "-of",
            "GTIFF",
            "-co",
            "TILED=YES",
            "-co",
            "COMPRESS=JPEG",
            "-co",
            "BIGTIFF=NO",
            "-ot",
            "Byte",
            "-r",
            "average",
            stretch_img,
            scaled_gtiff_img,
        ]

    try:
        subprocess.run(cmd, check=True)
    except OSError as e:
        raise rsgislib.RSGISPyException("Could not execute command: {}".format(cmd))
    rsgislib.imageutils.pop_img_stats(
        scaled_gtiff_img, use_no_data=True, no_data_val=0, calc_pyramids=True
    )
    shutil.rmtree(tmp_dir)


def create_quicklook_overview_imgs(
    input_imgs: List[str],
    bands: str,
    tmp_dir: str,
    output_imgs: Union[str, List[str]] = "quicklook.jpg",
    output_img_sizes: Union[int, List[int]] = 250,
    scale_axis: str = "auto",
    stretch_file: str = None,
    no_data_val: float = None,
):
    """
    A function to produce an overview quicklook for a number of input images.

    :param input_imgs: input image file (any format that gdal supports)
    :param bands: a string (comma seperated) with the bands to be selected.
                  (e.g., '1', '1,2,3', '5,6,4')
    :param tmp_dir: an input directory which can be used to write temporary
                    files / directories.
    :param output_imgs: a single output image or list of output images. The same
                        size as output_img_sizes. The output format is JPEG.
    :param output_img_sizes: the output image size (in pixels) or list of output
                             image sizes.
    :param scale_axis: the axis to which the output_img_sizes refer. Options: width,
                       height or auto. Auto applies the output_img_sizes to the
                       longest of the two axes.
    :param stretch_file: a stretch stats file to standardise the stretch between
                         a number of input files.
    :param no_data_val: Optionally provide an input image no data value otherwise
                        it will be read from the input image header.

    """
    if scale_axis not in ["width", "height", "auto"]:
        raise rsgislib.RSGISPyException(
            "Input parameter 'scale_axis' must have the value "
            "'width', 'height' or 'auto'."
        )

    if type(input_imgs) is not list:
        raise rsgislib.RSGISPyException("Input images must be a list")
    n_in_imgs = len(input_imgs)

    n_out_imgs = 1
    if type(output_imgs) is list:
        n_out_imgs = len(output_imgs)
        if type(output_img_sizes) is not list:
            raise rsgislib.RSGISPyException(
                "If the output_imgs input is a list so must output_img_sizes."
            )
        if len(output_img_sizes) != n_out_imgs:
            raise rsgislib.RSGISPyException(
                "output_imgs and output_img_sizes must be the same length"
            )

        if n_out_imgs == 1:
            output_imgs = output_imgs[0]
            output_img_sizes = output_img_sizes[0]

    uid_str = rsgislib.tools.utils.uid_generator()
    usr_tmp_dir = os.path.join(tmp_dir, "qklk_tmp_{}".format(uid_str))
    if not os.path.exists(usr_tmp_dir):
        os.makedirs(usr_tmp_dir)

    img_basename = rsgislib.tools.filetools.get_file_basename(
        input_imgs[0], check_valid=True
    )

    band_lst = bands.split(",")
    multi_band = False
    if len(band_lst) == 3:
        multi_band = True
    elif len(band_lst) == 1:
        multi_band = False
    else:
        raise rsgislib.RSGISPyException(
            "You need to either provide 1 or 3 bands: {}".format(band_lst)
        )

    s_bands = []
    for str_band in band_lst:
        s_bands.append(int(str_band))

    n_img_bands = rsgislib.imageutils.get_img_band_count(input_imgs[0])

    usr_def_no_data = False
    if no_data_val is None:
        no_data_val = rsgislib.imageutils.get_img_no_data_value(input_imgs[0])
    else:
        usr_def_no_data = True
    if no_data_val is None:
        raise rsgislib.RSGISPyException("No data value is not defined.")

    b_sel_imgs = []
    for img in input_imgs:
        tmp_n_bands = rsgislib.imageutils.get_img_band_count(img)
        tmp_no_data_val = rsgislib.imageutils.get_img_no_data_value(img)

        if tmp_n_bands != n_img_bands:
            raise rsgislib.RSGISPyException(
                "The number of bands in the input images is different."
            )

        if (not usr_def_no_data) and (tmp_no_data_val != no_data_val):
            raise rsgislib.RSGISPyException(
                "The no data value is different between the input images."
            )

        sel_img_bands_img = ""
        if (n_img_bands == 1) and (not multi_band):
            sel_img_bands_img = img
        elif (
            (n_img_bands == 3)
            and (multi_band)
            and (band_lst[0] == "1")
            and (band_lst[1] == "2")
            and (band_lst[2] == "3")
        ):
            sel_img_bands_img = img
        else:
            lcl_img_basename = rsgislib.tools.filetools.get_file_basename(
                img, check_valid=True
            )
            sel_img_bands_img = os.path.join(
                usr_tmp_dir, "{}_sband.kea".format(lcl_img_basename)
            )
            rsgislib.imageutils.select_img_bands(
                img,
                sel_img_bands_img,
                "KEA",
                rsgislib.imageutils.get_rsgislib_datatype_from_img(img),
                s_bands,
            )
            rsgislib.imageutils.set_img_no_data_value(sel_img_bands_img, no_data_val)
        b_sel_imgs.append(sel_img_bands_img)

    tmp_vrt_img = os.path.join(usr_tmp_dir, "{}_{}.vrt".format(img_basename, uid_str))
    rsgislib.imageutils.create_mosaic_images_vrt(b_sel_imgs, tmp_vrt_img)

    stretch_img = os.path.join(usr_tmp_dir, "{}_stretch.kea".format(img_basename))
    if stretch_file is not None:
        rsgislib.imageutils.stretch_img_with_stats(
            tmp_vrt_img,
            stretch_img,
            stretch_file,
            "KEA",
            rsgislib.TYPE_8UINT,
            no_data_val,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )
    else:
        rsgislib.imageutils.stretch_img(
            tmp_vrt_img,
            stretch_img,
            False,
            "",
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

    if scale_axis == "auto":
        x_size, y_size = rsgislib.imageutils.get_img_size(stretch_img)
        if x_size > y_size:
            scale_axis = "width"
        else:
            scale_axis = "height"

    if n_out_imgs == 1:
        if not rsgislib.tools.filetools.does_path_exists_or_creatable(output_imgs):
            raise rsgislib.RSGISPyException(
                f"Output file path is not creatable: {output_imgs}"
            )
        cmd = ["gdal_translate", "-of", "JPEG", "-ot", "Byte"]
        if scale_axis == "width":
            cmd.append("-outsize")
            cmd.append(f"{int(output_img_sizes)}")
            cmd.append("0")
        else:
            cmd.append("-outsize")
            cmd.append("0")
            cmd.append(f"{int(output_img_sizes)}")

        cmd.append("-r")
        cmd.append("average")
        cmd.append(stretch_img)
        cmd.append(output_imgs)

        print(cmd)
        try:
            subprocess.run(cmd, check=True)
        except OSError as e:
            raise rsgislib.RSGISPyException("Could not execute command: {}".format(cmd))

    elif n_out_imgs > 1:
        for i in range(n_out_imgs):
            if not rsgislib.tools.filetools.does_path_exists_or_creatable(
                output_imgs[i]
            ):
                raise rsgislib.RSGISPyException(
                    f"Output file path is not creatable: {output_imgs[i]}"
                )
            cmd = ["gdal_translate", "-of", "JPEG", "-ot", "Byte"]
            if scale_axis == "width":
                cmd.append("-outsize")
                cmd.append(f"{int(output_img_sizes[i])}")
                cmd.append("0")
            else:
                cmd.append("-outsize")
                cmd.append("0")
                cmd.append(f"{int(output_img_sizes[i])}")

            cmd.append("-r")
            cmd.append("average")
            cmd.append(stretch_img)
            cmd.append(output_imgs[i])

            print(cmd)
            try:
                subprocess.run(cmd, check=True)
            except OSError as e:
                raise rsgislib.RSGISPyException(
                    "Could not execute command: {}".format(cmd)
                )
    shutil.rmtree(usr_tmp_dir)


def burn_in_binary_msk(
    base_image: str, mask_img: str, output_img: str, gdalformat: str, msk_colour=None
):
    """
    A function which is used for visualisation applications where a rasterised
    binary mask is 'burnt' into a base image with a user selected colour.

    :param base_image: the base image with continuous output values. It is expected
                       that this image has 8uint pixel values.
    :param mask_img: the binary mask (value 1 will be used) to be 'burnt'
                     into the base_image.
    :param output_img: the output image file name and path
    :param gdalformat: the GDAL image file format of the output image file.
    :param msk_colour: the colour of the pixels burnt in to the base image. Should
                       have the same number of dimensions as the input base image
                       has image bands. Default, if None is 255.

    """
    from rios import applier
    import numpy

    n_img_bands = rsgislib.imageutils.get_img_band_count(base_image)
    if msk_colour is None:
        msk_colour = list()
        for i in range(n_img_bands):
            msk_colour.append(255)
    elif len(msk_colour) != n_img_bands:
        raise rsgislib.RSGISPyException(
            "The number of image bands and length of "
            "the msk_colour array should be equal."
        )

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    infiles = applier.FilenameAssociations()
    infiles.base_image = base_image
    infiles.mask_img = mask_img
    outfiles = applier.FilenameAssociations()
    outfiles.output_img = output_img
    otherargs = applier.OtherInputs()
    otherargs.msk_colour = msk_colour
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False

    def _burn_in_values(info, inputs, outputs, otherargs):
        """
        This is an internal rios function
        """
        outputs.output_img = numpy.copy(inputs.base_image)
        for n in range(inputs.base_image.shape[0]):
            outputs.output_img[n] = numpy.where(
                inputs.mask_img == 1, otherargs.msk_colour[n], outputs.output_img[n]
            )

    applier.apply(_burn_in_values, infiles, outfiles, otherargs, controls=aControls)


def create_quicklook_overview_imgs_vec_overlay(
    input_imgs: List[str],
    bands: str,
    tmp_dir: str,
    vec_overlay_file: str,
    vec_overlay_lyr: str,
    output_imgs: Union[str, List[str]] = "quicklook.jpg",
    output_img_sizes: Union[int, List[int]] = 250,
    gdalformat: str = "JPEG",
    scale_axis: str = "auto",
    stretch_file: str = None,
    overlay_clrs: List[int] = None,
    no_data_val: float = None,
):
    """
    A function to produce an overview quicklook with a vector overlain for
    a number of input images.

    :param input_imgs: input image file (any format that gdal supports)
    :param bands: a string (comma seperated) with the bands to be selected.
                  (e.g., '1', '1,2,3', '5,6,4')
    :param tmp_dir: an input directory which can be used to write temporary
                    files / directories.
    :param vec_overlay_file: an vector overlay which will be rasterised to the
                             overlay on the output images.
    :param vec_overlay_lyr: the layer name for the vector overlay.
    :param output_imgs: a single output image or list of output images. The same size
                        as output_img_sizes.
    :param output_img_sizes: the output image size (in pixels) or list of output
                             image sizes.
    :param gdalformat: the output file format - probably either JPG, PNG or GTIFF.
    :param scale_axis: the axis to which the output_img_sizes refer. Options: width,
                       height or auto. Auto applies the output_img_sizes to the
                       longest of the two axes.
    :param stretch_file: a stretch stats file to standardise the stretch between a
                         number of input files.
    :param overlay_clrs: list of output pixel values defining colour for the overlay
                         image (value between 0-255). The list should be the same
                         length as the number of input image bands (i.e., specified
                         by bands). If None then default all values are all 255.
    :param no_data_val: Optionally provide an input image no data value otherwise
                        it will be read from the input image header.

    """
    import rsgislib.vectorutils.createrasters

    if scale_axis not in ["width", "height", "auto"]:
        raise rsgislib.RSGISPyException(
            "Input parameter 'scale_axis' must have "
            "the value 'width', 'height' or 'auto'."
        )

    if type(input_imgs) is not list:
        raise rsgislib.RSGISPyException("Input images must be a list")
    n_in_imgs = len(input_imgs)

    n_out_imgs = 1
    if type(output_imgs) is list:
        n_out_imgs = len(output_imgs)
        if type(output_img_sizes) is not list:
            raise rsgislib.RSGISPyException(
                "If the outputImgs input is a list so must output_img_sizes."
            )
        if len(output_img_sizes) != n_out_imgs:
            raise rsgislib.RSGISPyException(
                "outputImgs and output_img_sizes must be the same length"
            )

        if n_out_imgs == 1:
            output_imgs = output_imgs[0]
            output_img_sizes = output_img_sizes[0]

    uid_str = rsgislib.tools.utils.uid_generator()
    usr_tmp_dir = os.path.join(tmp_dir, "qklk_tmp_{}".format(uid_str))
    if not os.path.exists(usr_tmp_dir):
        os.makedirs(usr_tmp_dir)

    img_basename = rsgislib.tools.filetools.get_file_basename(
        input_imgs[0], check_valid=True
    )
    band_lst = bands.split(",")
    multi_band = False
    if len(band_lst) == 3:
        multi_band = True
        out_n_bands = 3
    elif len(band_lst) == 1:
        multi_band = False
        out_n_bands = 1
    else:
        print(band_lst)
        raise rsgislib.RSGISPyException("You need to either provide 1 or 3 bands.")

    s_bands = []
    for strBand in band_lst:
        s_bands.append(int(strBand))

    n_img_bands = rsgislib.imageutils.get_img_band_count(input_imgs[0])
    usr_def_no_data = False
    if no_data_val is None:
        no_data_val = rsgislib.imageutils.get_img_no_data_value(input_imgs[0])
    else:
        usr_def_no_data = True
    if no_data_val is None:
        raise rsgislib.RSGISPyException("No data value is not defined.")

    b_sel_imgs = []
    for img in input_imgs:
        tmp_n_bands = rsgislib.imageutils.get_img_band_count(img)
        tmp_no_data_val = rsgislib.imageutils.get_img_no_data_value(img)

        if tmp_n_bands != n_img_bands:
            raise rsgislib.RSGISPyException(
                "The number of bands in the input images is different."
            )

        if (not usr_def_no_data) and (tmp_no_data_val != no_data_val):
            raise rsgislib.RSGISPyException(
                "The no data value is different between the input images."
            )

        sel_img_bands_img = ""
        if (n_img_bands == 1) and (not multi_band):
            sel_img_bands_img = img
        elif (
            (n_img_bands == 3)
            and (multi_band)
            and (band_lst[0] == "1")
            and (band_lst[1] == "2")
            and (band_lst[2] == "3")
        ):
            sel_img_bands_img = img
        else:
            lcl_img_basename = rsgislib.tools.filetools.get_file_basename(
                img, check_valid=True
            )
            sel_img_bands_img = os.path.join(
                usr_tmp_dir, "{}_sband.kea".format(lcl_img_basename)
            )
            rsgislib.imageutils.select_img_bands(
                img,
                sel_img_bands_img,
                "KEA",
                rsgislib.imageutils.get_rsgislib_datatype_from_img(img),
                s_bands,
            )
            rsgislib.imageutils.set_img_no_data_value(sel_img_bands_img, no_data_val)
        b_sel_imgs.append(sel_img_bands_img)

    if overlay_clrs is None:
        overlay_clrs = list()
        for i in range(out_n_bands):
            overlay_clrs.append(255)

    tmp_vrt_img = os.path.join(usr_tmp_dir, "{}_{}.vrt".format(img_basename, uid_str))
    rsgislib.imageutils.create_mosaic_images_vrt(b_sel_imgs, tmp_vrt_img)

    stretch_img = os.path.join(usr_tmp_dir, "{}_stretch.kea".format(img_basename))
    if stretch_file is not None:
        rsgislib.imageutils.stretch_img_with_stats(
            tmp_vrt_img,
            stretch_img,
            stretch_file,
            "KEA",
            rsgislib.TYPE_8UINT,
            no_data_val,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )
    else:
        rsgislib.imageutils.stretch_img(
            tmp_vrt_img,
            stretch_img,
            False,
            "",
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

    if scale_axis == "auto":
        x_size, y_size = rsgislib.imageutils.get_img_size(stretch_img)
        if x_size > y_size:
            scale_axis = "width"
        else:
            scale_axis = "height"

    if n_out_imgs == 1:
        cmd = ["gdal_translate", "-of", "KEA", "-ot", "Byte"]
        if scale_axis == "width":
            cmd.append("-outsize")
            cmd.append(f"{int(output_img_sizes)}")
            cmd.append("0")
        else:
            cmd.append("-outsize")
            cmd.append("0")
            cmd.append(f"{int(output_img_sizes)}")

        # Create the resized output image.
        lcl_img_basename = rsgislib.tools.filetools.get_file_basename(
            output_imgs, check_valid=True
        )
        tmp_resized_img = os.path.join(
            usr_tmp_dir, "{}_resized.kea".format(lcl_img_basename)
        )

        cmd.append("-r")
        cmd.append("average")
        cmd.append(stretch_img)
        cmd.append(tmp_resized_img)

        print(cmd)
        try:
            subprocess.run(cmd, check=True)
        except OSError as e:
            raise rsgislib.RSGISPyException("Could not execute command: {}".format(cmd))
        # Rasterise the overlay vector to the output raster grid.
        tmp_vec_overlay_img = os.path.join(
            usr_tmp_dir, "{}_vec_overlay.kea".format(lcl_img_basename)
        )
        rsgislib.vectorutils.createrasters.rasterise_vec_lyr(
            vec_overlay_file,
            vec_overlay_lyr,
            tmp_resized_img,
            tmp_vec_overlay_img,
            gdalformat="KEA",
            burn_val=1,
            datatype=rsgislib.TYPE_8UINT,
            att_column=None,
            thematic=True,
            no_data_val=0,
        )
        # Merge the overlay and base image
        tmp_final_img = os.path.join(
            usr_tmp_dir, "{}_final.kea".format(lcl_img_basename)
        )
        burn_in_binary_msk(
            tmp_resized_img, tmp_vec_overlay_img, tmp_final_img, "KEA", overlay_clrs
        )
        # Convert to final format (e.g., JPG, TIFF or PNG)
        if not rsgislib.tools.filetools.does_path_exists_or_creatable(output_imgs):
            raise rsgislib.RSGISPyException(
                f"Output file path is not creatable: {output_imgs}"
            )
        rsgislib.imageutils.gdal_translate(
            tmp_final_img, output_imgs, gdalformat=gdalformat, options=""
        )

    elif n_out_imgs > 1:
        for i in range(n_out_imgs):
            cmd = ["gdal_translate", "-of", "KEA", "-ot", "Byte"]
            if scale_axis == "width":
                cmd.append("-outsize")
                cmd.append(f"{int(output_img_sizes[i])}")
                cmd.append("0")
            else:
                cmd.append("-outsize")
                cmd.append("0")
                cmd.append(f"{int(output_img_sizes[i])}")

            # Create the resized output image.
            lcl_img_basename = rsgislib.tools.filetools.get_file_basename(
                output_imgs[i], check_valid=True
            )
            tmp_resized_img = os.path.join(
                usr_tmp_dir, "{}_resized.kea".format(lcl_img_basename)
            )
            cmd.append("-r")
            cmd.append("average")
            cmd.append(stretch_img)
            cmd.append(tmp_resized_img)

            print(cmd)
            try:
                subprocess.run(cmd, check=True)
            except OSError as e:
                raise rsgislib.RSGISPyException(
                    "Could not execute command: {}".format(cmd)
                )
            # Rasterise the overlay vector to the output raster grid.
            tmp_vec_overlay_img = os.path.join(
                usr_tmp_dir, "{}_vec_overlay.kea".format(lcl_img_basename)
            )
            rsgislib.vectorutils.createrasters.rasterise_vec_lyr(
                vec_overlay_file,
                vec_overlay_lyr,
                tmp_resized_img,
                tmp_vec_overlay_img,
                gdalformat="KEA",
                burn_val=1,
                datatype=rsgislib.TYPE_8UINT,
                att_column=None,
                thematic=True,
                no_data_val=0,
            )
            # Merge the overlay and base image
            tmp_final_img = os.path.join(
                usr_tmp_dir, "{}_final.kea".format(lcl_img_basename)
            )
            burn_in_binary_msk(
                tmp_resized_img, tmp_vec_overlay_img, tmp_final_img, "KEA", overlay_clrs
            )
            # Convert to final format (e.g., JPG, TIFF or PNG)
            if not rsgislib.tools.filetools.does_path_exists_or_creatable(
                output_imgs[i]
            ):
                raise rsgislib.RSGISPyException(
                    f"Output file path is not creatable: {output_imgs[i]}"
                )
            rsgislib.imageutils.gdal_translate(
                tmp_final_img, output_imgs[i], gdalformat=gdalformat, options=""
            )
    shutil.rmtree(usr_tmp_dir)


def create_visual_overview_imgs_vec_extent(
    input_imgs: List[str],
    bands: str,
    tmp_dir: str,
    vec_extent_file: str = None,
    vec_extent_lyr: str = None,
    output_imgs: Union[str, List[str]] = "quicklook.tif",
    output_img_sizes: Union[int, List[int]] = 500,
    gdalformat: str = "GTIFF",
    scale_axis: str = "auto",
    stretch_file: str = None,
    export_stretch_file: bool = False,
    no_data_val: float = None,
):
    """
    A function to produce an 8bit overview image (i.e., stretched visualisation)
    with an optional specified extent.

    :param input_imgs: input image file (any format that gdal supports)
    :param bands: a string (comma seperated) with the bands to be selected.
                  (e.g., '1', '1,2,3', '5,6,4')
    :param tmp_dir: an input directory which can be used to write temporary
                    files / directories.
    :param vec_extent_file: an vector file to define the extent of the output
                            image files.
    :param vec_extent_lyr: the layer name for the vector extent.
    :param output_imgs: a single output image or list of output images. The same
                        size as output_img_sizes.
    :param output_img_sizes: the output image size (in pixels) or list of output
                             image sizes.
    :param gdalformat: the output file format - probably either JPG, PNG or GTIFF.
    :param scale_axis: the axis to which the output_img_sizes refer. Options: width,
                       height or auto. Auto applies the output_img_sizes to the
                       longest of the two axes.
    :param stretch_file: a stretch stats file to standardise the stretch between a
                         number of input files. If export_stretch_file is True then
                         this variable is used as the output stretch file path.
    :param export_stretch_file: If true then the stretch parameters are outputted
                                as a text file to the path defined by stretch_file.
    :param no_data_val: Optionally provide an input image no data value otherwise
                        it will be read from the input image header.

    """
    import rsgislib.vectorutils.createvectors

    if (vec_extent_file is not None) and (vec_extent_lyr is None):
        raise rsgislib.RSGISPyException(
            "If a vec_extent_file is given (i.e., "
            "not None) then a vec_extent_lyr must also "
            "be provided."
        )

    if scale_axis not in ["width", "height", "auto"]:
        raise rsgislib.RSGISPyException(
            "Input parameter 'scale_axis' must have the value "
            "'width', 'height' or 'auto'."
        )

    if export_stretch_file and (stretch_file is None):
        raise rsgislib.RSGISPyException(
            "If export_stretch_file is True then a "
            "stretch_file path must be provided."
        )

    if type(input_imgs) is not list:
        raise rsgislib.RSGISPyException("Input images must be a list")
    n_in_imgs = len(input_imgs)

    n_out_imgs = 1
    if type(output_imgs) is list:
        n_out_imgs = len(output_imgs)
        if type(output_img_sizes) is not list:
            raise rsgislib.RSGISPyException(
                "If the output_imgs input is a list so must output_img_sizes."
            )
        if len(output_img_sizes) != n_out_imgs:
            raise rsgislib.RSGISPyException(
                "output_imgs and output_img_sizes must be the same length"
            )

        if n_out_imgs == 1:
            output_imgs = output_imgs[0]
            output_img_sizes = output_img_sizes[0]

    uid_str = rsgislib.tools.utils.uid_generator()
    usr_tmp_dir = os.path.join(tmp_dir, "qklk_tmp_{}".format(uid_str))
    if not os.path.exists(usr_tmp_dir):
        os.makedirs(usr_tmp_dir)

    img_basename = rsgislib.tools.filetools.get_file_basename(
        input_imgs[0], check_valid=True
    )
    band_lst = bands.split(",")
    multi_band = False
    if len(band_lst) == 3:
        multi_band = True
    elif len(band_lst) == 1:
        multi_band = False
    else:
        print(band_lst)
        raise rsgislib.RSGISPyException("You need to either provide 1 or 3 bands.")

    s_bands = []
    for str_band in band_lst:
        s_bands.append(int(str_band))

    n_img_bands = rsgislib.imageutils.get_img_band_count(input_imgs[0])

    usr_def_no_data = False
    if no_data_val is None:
        no_data_val = rsgislib.imageutils.get_img_no_data_value(input_imgs[0])
    else:
        usr_def_no_data = True
    if no_data_val is None:
        raise rsgislib.RSGISPyException("No data value is not defined.")

    b_sel_imgs = []
    for img in input_imgs:
        tmp_n_bands = rsgislib.imageutils.get_img_band_count(img)
        tmp_no_data_val = rsgislib.imageutils.get_img_no_data_value(img)

        if tmp_n_bands != n_img_bands:
            raise rsgislib.RSGISPyException(
                "The number of bands in the input images is different."
            )

        if (not usr_def_no_data) and (tmp_no_data_val != no_data_val):
            raise rsgislib.RSGISPyException(
                "The no data value is different between the input images."
            )

        sel_img_bands_img = ""
        if (n_img_bands == 1) and (not multi_band):
            sel_img_bands_img = img
        elif (
            (n_img_bands == 3)
            and (multi_band)
            and (band_lst[0] == "1")
            and (band_lst[1] == "2")
            and (band_lst[2] == "3")
        ):
            sel_img_bands_img = img
        else:
            lcl_img_basename = rsgislib.tools.filetools.get_file_basename(
                img, check_valid=True
            )
            sel_img_bands_img = os.path.join(
                usr_tmp_dir, "{}_sband.kea".format(lcl_img_basename)
            )
            rsgislib.imageutils.select_img_bands(
                img,
                sel_img_bands_img,
                "KEA",
                rsgislib.imageutils.get_rsgislib_datatype_from_img(img),
                s_bands,
            )
            rsgislib.imageutils.set_img_no_data_value(sel_img_bands_img, no_data_val)
        b_sel_imgs.append(sel_img_bands_img)

    tmp_vrt_img = os.path.join(usr_tmp_dir, "{}_{}.vrt".format(img_basename, uid_str))
    if vec_extent_file is None:
        rsgislib.imageutils.create_mosaic_images_vrt(b_sel_imgs, tmp_vrt_img)
    else:
        vec_extent_bbox = rsgislib.vectorutils.get_vec_layer_extent(
            vec_extent_file, vec_extent_lyr, compute_if_exp=True
        )
        vrt_extent = [
            vec_extent_bbox[0],
            vec_extent_bbox[2],
            vec_extent_bbox[1],
            vec_extent_bbox[3],
        ]
        rsgislib.imageutils.create_mosaic_images_vrt(
            b_sel_imgs, tmp_vrt_img, vrt_extent
        )

    stretch_img = os.path.join(usr_tmp_dir, "{}_stretch.kea".format(img_basename))
    if (stretch_file is not None) and (export_stretch_file == False):
        rsgislib.imageutils.stretch_img_with_stats(
            tmp_vrt_img,
            stretch_img,
            stretch_file,
            "KEA",
            rsgislib.TYPE_8UINT,
            no_data_val,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )
    else:
        rsgislib.imageutils.stretch_img(
            tmp_vrt_img,
            stretch_img,
            export_stretch_file,
            stretch_file,
            no_data_val,
            False,
            "KEA",
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )
    if scale_axis == "auto":
        x_size, y_size = rsgislib.imageutils.get_img_size(stretch_img)
        if x_size > y_size:
            scale_axis = "width"
        else:
            scale_axis = "height"

    if n_out_imgs == 1:
        if not rsgislib.tools.filetools.does_path_exists_or_creatable(output_imgs):
            raise rsgislib.RSGISPyException(
                f"Output file path is not creatable: {output_imgs}"
            )
        cmd = ["gdal_translate", "-of", gdalformat, "-ot", "Byte"]
        if scale_axis == "width":
            cmd.append("-outsize")
            cmd.append(f"{int(output_img_sizes)}")
            cmd.append("0")
        else:
            cmd.append("-outsize")
            cmd.append("0")
            cmd.append(f"{int(output_img_sizes)}")

        cmd.append("-r")
        cmd.append("average")
        cmd.append(stretch_img)
        cmd.append(output_imgs)

        print(cmd)
        try:
            subprocess.run(cmd, check=True)
        except OSError as e:
            raise rsgislib.RSGISPyException("Could not execute command: {}".format(cmd))
    else:
        for i in range(n_out_imgs):
            if not rsgislib.tools.filetools.does_path_exists_or_creatable(
                output_imgs[i]
            ):
                raise rsgislib.RSGISPyException(
                    f"Output file path is not creatable: {output_imgs[i]}"
                )
            cmd = ["gdal_translate", "-of", gdalformat, "-ot", "Byte"]
            if scale_axis == "width":
                cmd.append("-outsize")
                cmd.append(f"{int(output_img_sizes[i])}")
                cmd.append("0")
            else:
                cmd.append("-outsize")
                cmd.append("0")
                cmd.append(f"{int(output_img_sizes[i])}")

            cmd.append("-r")
            cmd.append("average")
            cmd.append(stretch_img)
            cmd.append(output_imgs[i])

            print(cmd)
            try:
                subprocess.run(cmd, check=True)
            except OSError as e:
                raise rsgislib.RSGISPyException(
                    "Could not execute command: {}".format(cmd)
                )
    shutil.rmtree(usr_tmp_dir)


def overlay_vec_on_img(
    input_img: str,
    output_img: str,
    vec_overlay_file: str,
    vec_overlay_lyr: str,
    tmp_dir: str,
    gdalformat: str = "PNG",
    overlay_clrs: List[int] = None,
):
    """
    A function to overlay a vector layer on to a raster image 'burning' in the vector
    as a particular colour. This is most appropriate for polyline vectors, polygons
    will be filled.

    :param input_img: The input image, usually a stretched byte (8uint) image ready
                      for visualisation but needs spatial header information associated.
    :param output_img: A output image file commonly in none specialist format such
                       as JPEG, PNG or TIFF.
    :param vec_overlay_file: The vector file to overlay (recommended to be a
                             polyline vector)
    :param vec_overlay_lyr: The name of the vector layer
    :param tmp_dir: A temporary were processing stage file can be written
                    during processing.
    :param gdalformat: The output format - commonly PNG, JPEG or GTIFF.
    :param overlay_clrs: An array with the output pixel values for the vector overlay,
                         needs the same length as the number of image bands. If None
                         then the overlay will be white (i.e., [255, 255, 255].

    """
    import rsgislib.vectorutils.createrasters

    # Check the overlay colour is defined and correct.
    n_img_bands = rsgislib.imageutils.get_img_band_count(input_img)
    if overlay_clrs is None:
        overlay_clrs = list()
        for i in range(n_img_bands):
            overlay_clrs.append(255)
    elif len(overlay_clrs) != n_img_bands:
        raise rsgislib.RSGISPyException(
            "The overlay colour does not have the same number of "
            "values as the number of bands within the input image."
        )

    # Create a temporary directory for processing stage outputs.
    img_basename = rsgislib.tools.filetools.get_file_basename(input_img)
    uid_str = rsgislib.tools.utils.uid_generator()
    usr_tmp_dir = os.path.join(
        tmp_dir, "{}_qklk_overlay_tmp_{}".format(img_basename, uid_str)
    )
    if not os.path.exists(usr_tmp_dir):
        os.makedirs(usr_tmp_dir)

    # Create raster of the vector to be overlain.
    tmp_vec_overlay_img = os.path.join(
        usr_tmp_dir, "{}_vec_overlay.kea".format(img_basename)
    )
    rsgislib.vectorutils.createrasters.rasterise_vec_lyr(
        vec_overlay_file,
        vec_overlay_lyr,
        input_img,
        tmp_vec_overlay_img,
        gdalformat="KEA",
        burn_val=1,
        datatype=rsgislib.TYPE_8UINT,
        att_column=None,
        thematic=True,
        no_data_val=0,
    )

    # Merge the overlay and base image
    tmp_final_img = os.path.join(usr_tmp_dir, "{}_final.kea".format(img_basename))
    burn_in_binary_msk(
        input_img, tmp_vec_overlay_img, tmp_final_img, "KEA", overlay_clrs
    )

    # Convert to final format (e.g., JPG, TIFF or PNG)
    rsgislib.imageutils.gdal_translate(tmp_final_img, output_img, gdalformat=gdalformat)

    # Remove the temporary directory.
    shutil.rmtree(usr_tmp_dir)



#!/usr/bin/env python
"""
The tools.google_cloud module contains functions for manipulating
"""
import os

import rsgislib


def blob_exists(goog_cred, bucket_name, filename):
    """
    A function to test whether a file .

    :param file_to_upload: The local file to be uploaded.
    :param goog_cred: The JSON credentials file from Google.
    :param bucket_name: The google bucket name.
    :param filename: The file path within the bucket for the file. Note this should not start
                     with a slash. i.e., directory/path rather than /directory/path.
    :return: boolean (True if the file exists / False if the file does not exist)
    """
    from google.cloud import storage

    client = storage.Client.from_service_account_json(goog_cred)
    bucket = client.get_bucket(bucket_name)
    blob = bucket.blob(filename)
    return blob.exists()


def list_files(goog_cred, bucket_name, bucket_dir):
    """
    A function to list the contents of a directory within a bucket.

    :param goog_cred: The JSON credentials file from Google.
    :param bucket_name: The google bucket name.
    :param bucket_dir: The directory path within the bucket to be listed. Note this
                       should not start with a slash. i.e., directory/path rather than /directory/path.
    :return: returns a list of file paths.

    """
    from google.cloud import storage

    client = storage.Client.from_service_account_json(goog_cred)
    bucket = client.get_bucket(bucket_name)
    files = bucket.list_blobs(prefix=bucket_dir)
    fileList = [file.name for file in files if "." in file.name]
    return fileList


def upload_to_google_bucket(file_to_upload, goog_cred, bucket_name, bucket_dir):
    """
    A function to upload a file to a google cloud bucket. Once the file is uploaded the presence
    of the file is checked. If the file is not present then an exception is raised.

    :param file_to_upload: The local file to be uploaded.
    :param goog_cred: The JSON credentials file from Google.
    :param bucket_name: The google bucket name.
    :param bucket_dir: The directory path within the bucket for the file to be saved. Note this
                       should not start with a slash. i.e., directory/path rather than /directory/path.

    """
    from google.cloud import storage

    # Explicitly use service account credentials by specifying the private key file.
    storage_client = storage.Client.from_service_account_json(goog_cred)

    lcl_path, file_name = os.path.split(file_to_upload)
    bucket_file_path = os.path.join(bucket_dir, file_name)
    bucket_obj = storage_client.get_bucket(bucket_name)
    blob_obj = bucket_obj.blob(bucket_file_path)
    blob_obj.upload_from_filename(file_to_upload)
    if not blob_exists(goog_cred, bucket_name, bucket_file_path):
        raise rsgislib.RSGISPyException(
            "File was not successfully uploaded to Google Cloud: '{}'".format(
                file_to_upload
            )
        )



import os
import pytest

GEOPANDAS_NOT_AVAIL = False
try:
    import geopandas
except ImportError:
    GEOPANDAS_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def test_polygonise_raster_to_vec_lyr_4_conn(tmp_path):
    import rsgislib.vectorutils.createvectors

    input_img = os.path.join(DATA_DIR, "aber_osgb_multi_polys_rasters.kea")

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.polygonise_raster_to_vec_lyr(
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        input_img=input_img,
        img_band=1,
        mask_img=input_img,
        mask_band=1,
        use_8_conn=False,
    )

    assert os.path.exists(out_vec_file)


def test_polygonise_raster_to_vec_lyr_8_conn(tmp_path):
    import rsgislib.vectorutils.createvectors

    input_img = os.path.join(DATA_DIR, "aber_osgb_multi_polys_rasters.kea")

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.polygonise_raster_to_vec_lyr(
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        input_img=input_img,
        img_band=1,
        mask_img=input_img,
        mask_band=1,
        use_8_conn=True,
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_vectorise_pxls_to_pts(tmp_path):
    import rsgislib.vectorutils.createvectors

    input_img = os.path.join(DATA_DIR, "aber_osgb_multi_polys_rasters.kea")
    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.vectorise_pxls_to_pts(
        input_img,
        1,
        3,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        out_epsg_code=None,
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_extract_image_footprint(tmp_path):
    import rsgislib.vectorutils.createvectors

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.extract_image_footprint(
        input_img,
        out_vec_file,
        out_vec_lyr,
        "GPKG",
        tmp_dir=tmp_path,
        reproj_to=None,
        no_data_val=0,
    )

    assert os.path.exists(out_vec_file)


def test_extract_image_footprint_exception(tmp_path):
    import rsgislib.vectorutils.createvectors

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    with pytest.raises(rsgislib.RSGISPyException):
        rsgislib.vectorutils.createvectors.extract_image_footprint(
            input_img,
            out_vec_file,
            out_vec_lyr,
            "GPKG",
            tmp_dir=tmp_path,
            reproj_to=None,
        )


def test_extract_image_footprint_reproj(tmp_path):
    import rsgislib.vectorutils.createvectors

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.extract_image_footprint(
        input_img,
        out_vec_file,
        out_vec_lyr,
        "GPKG",
        tmp_dir=tmp_path,
        reproj_to="epsg:4326",
        no_data_val=0,
    )

    assert os.path.exists(out_vec_file)


def test_define_grid(tmp_path):
    import rsgislib.vectorutils.createvectors

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.define_grid(
        [0, 5, 10, 15],
        1,
        1,
        4326,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        out_epsg_code=None,
        utm_grid=False,
        utm_hemi=False,
    )

    assert os.path.exists(out_vec_file)


def test_define_grid_utm(tmp_path):
    import rsgislib.vectorutils.createvectors

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.define_grid(
        [0, 5, 10, 15],
        5000,
        5000,
        4326,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        out_epsg_code=None,
        utm_grid=True,
        utm_hemi=False,
    )

    assert os.path.exists(out_vec_file)


def test_create_poly_vec_bboxs(tmp_path):
    import rsgislib.vectorutils.createvectors
    from osgeo import ogr

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    bboxs = [[1, 2, 1, 2], [2, 3, 2, 3], [3, 4, 3, 4], [4, 5, 4, 5]]

    atts = dict()
    atts["name"] = ["hello", "to", "the", "world"]
    atts["id"] = [1, 2, 3, 4]

    att_types = dict()
    att_types["names"] = ["name", "id"]
    att_types["types"] = [ogr.OFTString, ogr.OFTInteger]

    rsgislib.vectorutils.createvectors.create_poly_vec_bboxs(
        out_vec_file,
        out_vec_lyr,
        "GPKG",
        4326,
        bboxs,
        atts=atts,
        att_types=att_types,
        overwrite=True,
    )

    assert os.path.exists(out_vec_file)


def test_write_pts_to_vec(tmp_path):
    import rsgislib.vectorutils.createvectors
    from osgeo import ogr

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    pts_x = [1, 2, 3, 4]
    pts_y = [1, 2, 3, 4]

    atts = dict()
    atts["name"] = ["hello", "to", "the", "world"]
    atts["id"] = [1, 2, 3, 4]

    att_types = dict()
    att_types["names"] = ["name", "id"]
    att_types["types"] = [ogr.OFTString, ogr.OFTInteger]

    rsgislib.vectorutils.createvectors.write_pts_to_vec(
        out_vec_file,
        out_vec_lyr,
        "GPKG",
        4326,
        pts_x,
        pts_y,
        atts=atts,
        att_types=att_types,
        replace=True,
        file_opts=[],
        lyr_opts=[],
    )

    assert os.path.exists(out_vec_file)


def test_create_bboxs_for_pts(tmp_path):
    import rsgislib.vectorutils.createvectors

    pts_vec_file = os.path.join(tmp_path, "pts_vec.gpkg")
    pts_vec_lyr = "pts_vec"

    pts_x = [1, 2, 3, 4]
    pts_y = [1, 2, 3, 4]

    rsgislib.vectorutils.createvectors.write_pts_to_vec(
        pts_vec_file,
        pts_vec_lyr,
        "GPKG",
        4326,
        pts_x,
        pts_y,
        atts=None,
        att_types=None,
        replace=True,
        file_opts=[],
        lyr_opts=[],
    )

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.create_bboxs_for_pts(
        pts_vec_file,
        pts_vec_lyr,
        0.25,
        0.35,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        del_exist_vec=False,
    )
    assert os.path.exists(out_vec_file)


def test_create_wgs84_vector_grid(tmp_path):
    import rsgislib.vectorutils.createvectors

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.create_wgs84_vector_grid(out_vec_file, out_vec_lyr, 'GPKG', 1, 1, [
        -5, 5, -5, 5], overlap=None, tile_name_prefix='test_')

    assert os.path.exists(out_vec_file)


def test_create_wgs84_vector_grid_overlap(tmp_path):
    import rsgislib.vectorutils.createvectors

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorutils.createvectors.create_wgs84_vector_grid(out_vec_file, out_vec_lyr, 'GPKG', 1, 1, [
        -5, 5, -5, 5], overlap=0.0277777778, tile_name_prefix='test_')

    assert os.path.exists(out_vec_file)



#!/usr/bin/env python
"""
The tools.plotting module contains functions for extracting and plotting remote sensing data.
"""

from typing import Tuple, List, Dict, Union
import os

import rsgislib
from rsgislib import zonalstats
from rsgislib import imagecalc

from osgeo import gdal
import numpy

have_matplotlib = True
try:
    import matplotlib.pyplot as plt
    import matplotlib.colors as mClrs
    from matplotlib.patches import Patch
    from matplotlib import rcParams
    from matplotlib import gridspec
except ImportError:
    have_matplotlib = False


def plot_image_spectra(
    input_img,
    vec_file,
    vec_lyr,
    output_plot_file,
    wavelengths,
    plot_title,
    scale_factor=0.1,
    show_refl_std=True,
    refl_max=None,
):
    """
    A utility function to extract and plot image spectra.

    :param input_img: is the input image
    :param vec_file: is the region of interest file as a vector file - if multiple polygons are defined the spectra for each will be added to the plot.
    :param vec_lyr:
    :param output_plot_file: is the output PDF file for the plot which has been create
    :param wavelengths: is list of numbers with the wavelength of each band (must have the same number of wavelengths as image bands)
    :param plot_title: is a string with the title for the plot
    :param scale_factor: is a float specifying the scaling to percentage (0 - 100). (Default is 0.1, i.e., pixel values are scaled between 0-1000; ARCSI default).
    :param show_refl_std: is a boolean (default: True) to specify whether a shaded region showing 1 standard deviation from the mean on the plot alongside the mean spectra.
    :param refl_max: is a parameter for setting the maximum reflectance value on the Y axis (if None the maximum value in the dataset is used

    .. code:: python

        from rsgislib.tools import plotting

        inputImage = 'injune_p142_casi_sub_utm.kea'
        roiFile = 'spectraROI.shp'
        outputPlotFile = 'SpectraPlot.pdf'
        wavelengths = [446.0, 530.0, 549.0, 569.0, 598.0, 633.0, 680.0, 696.0, 714.0, 732.0, 741.0, 752.0, 800.0, 838.0]
        plotTitle = "Image Spectral from CASI Image"

        plotting.plot_image_spectra(inputImage, roiFile, outputPlotFile, wavelengths, plotTitle)

    """
    import rsgislib.tools.filetools

    if not have_matplotlib:
        raise rsgislib.RSGISPyException(
            "The matplotlib module is required and could not be imported."
        )

    dataset = gdal.Open(input_img, gdal.GA_ReadOnly)
    numBands = dataset.RasterCount
    dataset = None

    if not len(wavelengths) == numBands:
        raise rsgislib.RSGISPyException(
            "The number of wavelengths and image bands must be equal."
        )

    tmpOutFile = os.path.splitext(output_plot_file)[0] + "_statstmp.txt"
    zonalattributes = zonalstats.ZonalAttributes(
        minThreshold=0,
        maxThreshold=10000,
        calcCount=False,
        calcMin=False,
        calcMax=False,
        calcMean=True,
        calcStdDev=True,
        calcMode=False,
        calcSum=False,
    )
    zonalstats.pixelStats2TXT(
        input_img,
        vec_file,
        tmpOutFile,
        zonalattributes,
        True,
        True,
        zonalstats.METHOD_POLYCONTAINSPIXELCENTER,
        False,
    )

    meanVals = []
    stdDevVals = []
    stats = open(tmpOutFile, "r")
    row = 0
    for statsRow in stats:
        statsRow = statsRow.strip()
        if row > 0:
            meanVal = []
            stdDevVal = []
            data = statsRow.split(",")

            if not len(data) == (numBands * 2) + 2:
                raise rsgislib.RSGISPyException(
                    "The number of outputted stats values is incorrect!"
                )
            for band in range(numBands):
                meanVal.append(float(data[(band * 2) + 1]) * scale_factor)
                stdDevVal.append(float(data[(band * 2) + 2]) * scale_factor)
            meanVals.append(meanVal)
            stdDevVals.append(stdDevVal)
        row += 1
    stats.close()

    # print("Mean: ", meanVals)
    # print("Std Dev: ", stdDevVals)
    print("Creating Plot")
    fig = plt.figure(figsize=(7, 5), dpi=80)
    ax1 = fig.add_subplot(111)
    for feat in range(len(meanVals)):
        ax1.plot(wavelengths, meanVals[feat], "k-", zorder=10)
        if show_refl_std:
            lowerVals = []
            upperVals = []
            for band in range(numBands):
                lowerVals.append(meanVals[feat][band] - stdDevVals[feat][band])
                upperVals.append(meanVals[feat][band] + stdDevVals[feat][band])
            ax1.fill_between(
                wavelengths,
                lowerVals,
                upperVals,
                alpha=0.2,
                linewidth=1.0,
                facecolor=[0.70, 0.70, 0.70],
                edgecolor=[0.70, 0.70, 0.70],
                zorder=-1,
            )

    ax1Range = ax1.axis("tight")

    if refl_max is None:
        ax1.axis((ax1Range[0], ax1Range[1], 0, ax1Range[3]))
    else:
        ax1.axis((ax1Range[0], ax1Range[1], 0, refl_max))

    plt.grid(color="k", linestyle="--", linewidth=0.5)
    plt.title(plot_title)
    plt.xlabel("Wavelength")
    plt.ylabel("Reflectance (%)")

    plt.savefig(output_plot_file, format="PDF")
    rsgislib.tools.filetools.delete_file_silent(tmpOutFile)
    print("Completed.\n")


def plot_image_comparison(
    inputImage1,
    inputImage2,
    img1Band,
    img2Band,
    outputPlotFile,
    numBins=100,
    img1Min=None,
    img1Max=None,
    img2Min=None,
    img2Max=None,
    img1Scale=1,
    img2Scale=1,
    img1Off=0,
    img2Off=0,
    normOutput=False,
    plotTitle="2D Histogram",
    xLabel="X Axis",
    yLabel="Y Axis",
    ctable="jet",
    interp="nearest",
):
    """
    A function to plot two images against each other.

    :param inputImage1: is a string with the path to the first image.
    :param inputImage2: is a string with the path to the second image.
    :param img1Band: is an int specifying the band in the first image to be plotted.
    :param img2Band: is an int specifying the band in the second image to be plotted.
    :param outputPlotFile: is a string specifying the output PDF for the plot.
    :param numBins: is an int specifying the number of bins within each axis of the histogram (default: 100)
    :param img1Min: is a double specifying the minimum value to be used in the histogram from image 1. If value is None then taken from the image.
    :param img1Max: is a double specifying the maximum value to be used in the histogram from image 1. If value is None then taken from the image.
    :param img2Min: is a double specifying the minimum value to be used in the histogram from image 2. If value is None then taken from the image.
    :param img2Max: is a double specifying the maximum value to be used in the histogram from image 2. If value is None then taken from the image.
    :param img1Scale: is a double specifying the scale for image 1 (Default 1).
    :param img2Scale: is a double specifying the scale for image 2 (Default 1).
    :param img1Off: is a double specifying the offset for image 1 (Default 0).
    :param img2Off: is a double specifying the offset for image 2 (Default 0).
    :param normOutput: is a boolean specifying whether the histogram should be normalised (Default: False).
    :param plotTitle: is a string specifying the title of the plot (Default: '2D Histogram').
    :param xLabel: is a string specifying the x axis label (Default: 'X Axis')
    :param yLabel: is a string specifying the y axis label (Default: 'Y Axis')
    :param ctable: is a string specifying the colour table to be used (Default: jet), list of available colour tables specified by matplotlib: http://matplotlib.org/examples/color/colormaps_reference.html
    :param interp: is a string specifying the interpolation algorithm to be used (Default: 'nearest'). Available values are ‘none’, ‘nearest’, ‘bilinear’, ‘bicubic’, ‘spline16’, ‘spline36’, ‘hanning’, ‘hamming’, ‘hermite’, ‘kaiser’, ‘quadric’, ‘catrom’, ‘gaussian’, ‘bessel’, ‘mitchell’, ‘sinc’, ‘lanczos’.

    .. code:: python

        from rsgislib.tools import plotting

        inputImage1 = 'LS5TM_20000613_lat10lon6217_r67p231_rad_sref_ndvi.kea'
        inputImage2 = 'LS5TM_20000613_lat10lon6217_r67p231_rad_ndvi.kea'
        outputPlotFile = 'ARCSI_RAD_SREF_NDVI.pdf'

        plotting.plot_image_comparison(inputImage1, inputImage2, 1, 1, outputPlotFile, img1Min=-0.5, img1Max=1, img2Min=-0.5, img2Max=1, plotTitle='ARCSI SREF NDVI vs ARCSI RAD NDVI', xLabel='ARCSI SREF NDVI', yLabel='ARCSI RAD NDVI')

    """
    if not have_matplotlib:
        raise rsgislib.RSGISPyException(
            "The matplotlib module is required and could not be imported."
        )

    gdalformat = "KEA"
    tmpOutFile = os.path.splitext(outputPlotFile)[0] + "_hist2dimgtmp.kea"
    # tmpOutFileStch = os.path.splitext(outputPlotFile)[0] + "_hist2dimgtmpStch.kea"

    if (img1Min is None) or (img1Max is None):
        # Calculate image 1 stats
        imgGDALDS = gdal.Open(inputImage1, gdal.GA_ReadOnly)
        imgGDALBand = imgGDALDS.GetRasterBand(img1Band)
        min, max = imgGDALBand.ComputeRasterMinMax(False)
        imgGDALDS = None
        if img1Min is None:
            img1Min = min
        if img1Max is None:
            img1Max = max

    if (img2Min is None) or (img2Max is None):
        # Calculate image 2 stats
        imgGDALDS = gdal.Open(inputImage2, gdal.GA_ReadOnly)
        imgGDALBand = imgGDALDS.GetRasterBand(img2Band)
        min, max = imgGDALBand.ComputeRasterMinMax(False)
        imgGDALDS = None
        if img2Min is None:
            img2Min = min
        if img2Max is None:
            img2Max = max

    # Images are flipped so axis' come out correctly.
    outBinSizeImg1, outBinSizeImg2, rSq = imagecalc.get2DImageHistogram(
        inputImage2,
        inputImage1,
        tmpOutFile,
        gdalformat,
        img2Band,
        img1Band,
        numBins,
        img2Min,
        img2Max,
        img1Min,
        img1Max,
        img2Scale,
        img1Scale,
        img2Off,
        img1Off,
        normOutput,
    )
    print("Image1 Bin Size: ", outBinSizeImg1)
    print("Image2 Bin Size: ", outBinSizeImg2)
    print("rSq: ", rSq)

    print("Read Image Data")
    plotGDALImg = gdal.Open(tmpOutFile, gdal.GA_ReadOnly)
    plotImgBand = plotGDALImg.GetRasterBand(1)
    dataArr = plotImgBand.ReadAsArray().astype(float)
    plotGDALImg = None

    fig = plt.figure(figsize=(7, 7), dpi=80)
    ax1 = fig.add_subplot(111)

    img1MinSc = img1Off + (img1Min * img1Scale)
    img1MaxSc = img1Off + (img1Max * img1Scale)
    img2MinSc = img2Off + (img2Min * img2Scale)
    img2MaxSc = img2Off + (img2Max * img2Scale)

    minVal = numpy.min(dataArr[dataArr != 0])
    maxVal = numpy.max(dataArr)

    print("Min Value: ", minVal)
    print("Max Value: ", maxVal)

    cmap = plt.get_cmap(ctable)
    mClrs.Colormap.set_under(cmap, color="white")
    mClrs.Colormap.set_over(cmap, color="white")

    imPlot = plt.imshow(
        dataArr,
        cmap=cmap,
        aspect="equal",
        interpolation=interp,
        norm=mClrs.Normalize(vmin=minVal, vmax=maxVal),
        vmin=minVal,
        vmax=maxVal,
        origin=[0, 0],
        extent=[img1MinSc, img1MaxSc, img2MinSc, img2MaxSc],
    )
    plt.grid(color="k", linestyle="--", linewidth=0.5)
    rSqStr = ""
    if rSq < 0:
        rSq = 0.00
    rSqStr = "$r^2 = " + str(round(rSq, 3)) + "$"
    plt.text(0.05, 0.95, rSqStr, va="center", transform=ax1.transAxes)
    fig.colorbar(imPlot)
    plt.title(plotTitle)
    plt.xlabel(xLabel)
    plt.ylabel(yLabel)

    plt.savefig(outputPlotFile, format="PDF")

    # Tidy up temporary file.
    gdalDriver = gdal.GetDriverByName(gdalformat)
    gdalDriver.Delete(tmpOutFile)


def plot_image_histogram(
    input_img,
    imgBand,
    outputPlotFile,
    numBins=100,
    imgMin=None,
    imgMax=None,
    normOutput=False,
    plotTitle="Histogram",
    xLabel="X Axis",
    colour="blue",
    edgecolour="black",
    linewidth=None,
):
    """
    A function to plot the histogram of an image.

    :param input_img: is a string with the path to the image.
    :param imgBand: is an int specifying the band in the image to be plotted.
    :param outputPlotFile: is a string specifying the output PDF for the plot.
    :param numBins: is an int specifying the number of bins within each axis of the histogram (default: 100)
    :param imgMin: is a double specifying the minimum value to be used in the histogram from the image. If value is None then taken from the image.
    :param imgMax: is a double specifying the maximum value to be used in the histogram from the image. If value is None then taken from the image.
    :param normOutput: is a boolean specifying whether the histogram should be normalised (Default: False).
    :param plotTitle: is a string specifying the title of the plot (Default: '2D Histogram').
    :param xLabel: is a string specifying the x axis label (Default: 'X Axis')
    :param colour: is the colour of the bars in the plot (see matplotlib documentation for how to specify, either keyword or RGB values (e.g., [1.0,0,0])
    :param edgecolour: is the colour of the edges of the bars
    :param linewidth: is the thickness of the edges of the bars in the plot.

    .. code:: python

        from rsgislib.tools import plotting

        plotting.plot_image_histogram("Baccini_Manaus_AGB_30.kea", 1, "BacciniHistogram.pdf", numBins=100, imgMin=0, imgMax=400, normOutput=True, plotTitle='Histogram of Baccini Biomass', xLabel='Baccini Biomass', color=[1.0,0.2,1.0], edgecolor='red', linewidth=0)

    """
    if not have_matplotlib:
        raise rsgislib.RSGISPyException(
            "The matplotlib module is required and could not be imported."
        )

    if (imgMin is None) or (imgMax is None):
        # Calculate image 1 stats
        imgGDALDS = gdal.Open(input_img, gdal.GA_ReadOnly)
        imgGDALBand = imgGDALDS.GetRasterBand(imgBand)
        min, max = imgGDALBand.ComputeRasterMinMax(False)
        imgGDALDS = None
        if imgMin is None:
            imgMin = min
        if imgMax is None:
            imgMax = max

    binWidth = (imgMax - imgMin) / numBins
    print("Bin Size: ", binWidth)

    bins, hMin, hMax = imagecalc.getHistogram(
        input_img, imgBand, binWidth, False, imgMin, imgMax
    )

    if normOutput:
        sumBins = numpy.sum(bins)
        bins = bins / sumBins

    numBins = len(bins)
    xLocs = numpy.arange(numBins)
    xLocs = (xLocs * binWidth) - (binWidth / 2)

    fig = plt.figure(figsize=(7, 7), dpi=80)
    plt.bar(
        xLocs,
        bins,
        width=binWidth,
        color=colour,
        edgecolor=edgecolour,
        linewidth=linewidth,
    )
    plt.xlim(imgMin, imgMax)

    plt.title(plotTitle)
    plt.xlabel(xLabel)
    plt.ylabel("Freq.")
    plt.savefig(outputPlotFile, format="PDF")


def residual_plot(y_true, residuals, out_file, out_format="PNG", title=None):
    """
    A function to create a residual plot to investigate the
    normality and homoscedasticity of model residuals.

    :param y_true: A numpy 1D array containing true/observed values.
    :param residuals: A numpy 1D array containing model residuals.
    :param out_file: Path to the output file.
    :param out_format: Output format supported by matplotlib (e.g. "PNG" or "PDF").
                       Default: PNG
    :param title: A title for the plot. Optional, if None then ignored. (Default: None)

    """
    if not have_matplotlib:
        raise rsgislib.RSGISPyException(
            "The matplotlib module is required and could not be imported."
        )

    if not isinstance(residuals, numpy.ndarray):
        residuals = numpy.array(residuals)
    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.array(y_true)
    if y_true.ndim != 1:
        raise rsgislib.RSGISPyException("y_true has more than 1 dimension.")
    if residuals.ndim != 1:
        raise rsgislib.RSGISPyException("Residuals has more than 1 dimension.")
    if residuals.size != y_true.size:
        raise rsgislib.RSGISPyException("y_true.size != residuals.size.")

    # setup plot:
    # rcParams.update({'font.family': 'cmr10'})  # use latex fonts.
    # rcParams['axes.unicode_minus'] = False
    rcParams.update({"font.size": 8.5})
    rcParams["axes.linewidth"] = 0.5
    rcParams["xtick.major.pad"] = "2"
    rcParams["ytick.major.pad"] = "2"
    fig = plt.figure(figsize=(5, 2.5))
    gs = gridspec.GridSpec(nrows=1, ncols=2, width_ratios=[3.5, 1])
    ax1 = plt.subplot(gs[0])
    ax2 = plt.subplot(gs[1])
    plt.tight_layout(w_pad=-1, h_pad=0)

    # draw scatterplot:
    ax1.axhline(y=0.0, c="k", ls=":", lw=0.5, zorder=2)
    ax1.scatter(y_true, residuals, s=16, color="C0", marker=".", linewidth=0, zorder=1)
    ax1.set_xlabel("Observed value", fontsize=9)
    ax1.set_ylabel("Residuals", fontsize=9)
    if title is not None:
        ax1.set_title(title)

    # draw histogram:
    ax2.get_xaxis().tick_bottom()
    ax2.get_yaxis().tick_right()
    ax2.get_yaxis().set_visible(False)
    ax2.hist(residuals, bins=50, orientation="horizontal", color="C0")
    ax2.axhline(y=0.0, c="k", ls=":", lw=0.5, zorder=2)
    ax2.set_xlabel("Frequency", fontsize=9)
    plt.savefig(out_file, format=out_format, dpi=300, bbox_inches="tight")
    plt.close()


def quantile_plot(residuals, ylabel, out_file, out_format="PNG", title=None):
    """
    A function to create a Quantile-Quantile plot to investigate the
    normality of model residuals.

    :param residuals: A numpy 1D array containing model residuals.
    :param ylabel: A string defining a label for the y axis
    :param out_file: Path to the output file.
    :param out_format: Output format supported by matplotlib (e.g. "PNG" or "PDF").
                       Default: PNG
    :param title: A title for the plot. Optional, if None then ignored. (Default: None)

    """
    if not have_matplotlib:
        raise rsgislib.RSGISPyException(
            "The matplotlib module is required and could not be imported."
        )
    from scipy.stats import probplot

    if not isinstance(residuals, numpy.ndarray):
        residuals = numpy.array(residuals)

        # rcParams.update({'font.family': 'cmr10'})  # use latex fonts.
    # rcParams['axes.unicode_minus'] = False
    rcParams.update({"font.size": 8.5})
    rcParams["axes.linewidth"] = 0.5
    rcParams["xtick.major.pad"] = "2"
    rcParams["ytick.major.pad"] = "2"
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(3.5, 2.5))
    ax.get_yaxis().set_tick_params(which="major", direction="out")
    ax.get_xaxis().set_tick_params(which="major", direction="out")
    ax.get_xaxis().set_tick_params(which="minor", direction="out", length=0, width=0)
    ax.get_yaxis().set_tick_params(which="minor", direction="out", length=0, width=0)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.xaxis.set_tick_params(width=0.5)
    ax.yaxis.set_tick_params(width=0.5)

    probplot(residuals, dist="norm", plot=ax)

    if title is not None:
        ax.set_title(title)
    else:
        ax.set_title("")
    ax.set_ylabel(ylabel)
    ax.get_lines()[0].set_marker(".")
    ax.get_lines()[0].set_markerfacecolor("k")
    ax.get_lines()[0].set_markeredgecolor("k")
    ax.get_lines()[0].set_markeredgewidth(0)
    ax.get_lines()[0].set_markersize(4.0)
    ax.get_lines()[1].set_linewidth(1.0)
    ax.get_lines()[1].set_color("r")
    plt.savefig(out_file, format=out_format, dpi=300, bbox_inches="tight")
    plt.close()


def plot_histogram_threshold(
    data: numpy.array,
    out_file: str,
    threshold: float = None,
    title_str: str = None,
    hist_clr: str = "gray",
    out_format: str = None,
    plot_dpi: int = 300,
):
    """
    A function which creates a histogram plot for a numpy array of data and optionally
    displaying a threshold on the histogram

    :param data: The numpy array to be plotted. The array should be a 1d array.
    :param out_file: The output image file path for the plot
    :param threshold: An optional threshold to be plotted. If None then ignored.
    :param title_str: An optional string for the plot title.
    :param hist_clr: the colour for the histogram bars.
    :param out_format: the output format of the plot. If None then matplotlib uses
                       the output file extension.
    :param plot_dpi: the dpi for the output plot.

    """
    import rsgislib

    if not have_matplotlib:
        raise rsgislib.RSGISPyException(
            "The matplotlib module is required and could not be imported."
        )
    import rsgislib.tools.stats

    n_bins, bin_width = rsgislib.tools.stats.get_nbins_histogram(data)

    plt.figure()
    plt.hist(data, bins=n_bins, color=hist_clr)
    if threshold is not None:
        plt.axvline(x=threshold, color="red")
    if title_str is not None:
        plt.title(title_str)
    plt.savefig(out_file, format=out_format, dpi=plot_dpi)


def get_gdal_raster_mpl_imshow(
    input_img: str, bands: List[int] = None, bbox: List[float] = None
) -> Tuple[numpy.array, List[float]]:
    """
    A function which retrieves image data as an array in an appropriate structure
    for use within the matplotlib imshow function. The extent is also returned.
    Note, this function assumes that the image pixels values are within an appropriate
    range for display.

    :param input_img: The input image file path.
    :param bands: Optional list of image bands to be selected and returned. If not
                  provided then all bands will be read. However, note that only 3
                  or 1 band(s) are valid for visualisation and an error will be thrown
                  if the number of bands is not 3 or 1.
    :param bbox: Optional bbox (xmin, xmax, ymin, ymax) used to subset the
                 input image so only data for the subset are returned.
    :return: numpy.array either [n,m,3] or [n,m] and a bbox (xmin, xmax, ymin, ymax)
             specifying the extent of the image data.

    .. code:: python

        img_sub_bbox = [554756, 577168, 9903924, 9944315]
        input_img = "sen2_img_strch.kea"

        img_data_arr, coords_bbox = get_gdal_raster_mpl_imshow(input_img,
                                                               bands=[8,9,3],
                                                               bbox=img_sub_bbox)


        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        im = ax.imshow(img_data_arr, extent=coords_bbox)
        plt.show()

    """
    import rsgislib.imageutils

    n_bands = rsgislib.imageutils.get_img_band_count(input_img)
    ref_bands = list(range(1, n_bands + 1, 1))
    if bands is None:
        bands = ref_bands
    else:
        for band in bands:
            if band not in ref_bands:
                raise rsgislib.RSGISPyException(
                    f"Band {band} is not valid (i.e., within the image)"
                )

    if not (len(bands) == 3 or len(bands) == 1):
        raise rsgislib.RSGISPyException("The number of bands specified must be 3 or 1.")

    x_size, y_size = rsgislib.imageutils.get_img_size(input_img)
    pxl_bbox = [0, x_size, 0, y_size]
    if bbox is not None:
        pxl_bbox = rsgislib.imageutils.get_img_subset_pxl_bbox(input_img, bbox)

    n_x_pxls = pxl_bbox[1] - pxl_bbox[0]
    n_y_pxls = pxl_bbox[3] - pxl_bbox[2]

    print(f"Image Data Size: {n_x_pxls} x {n_y_pxls}")

    coords_bbox = rsgislib.imageutils.get_img_pxl_spatial_coords(input_img, pxl_bbox)

    image_ds = gdal.Open(input_img, gdal.GA_ReadOnly)
    if image_ds is None:
        raise rsgislib.RSGISPyException(
            f"Could not open the input image file: '{input_img}'"
        )

    img_data_arr = image_ds.ReadAsArray(
        xoff=pxl_bbox[0],
        yoff=pxl_bbox[2],
        xsize=n_x_pxls,
        ysize=n_y_pxls,
        band_list=bands,
    )
    image_ds = None
    if len(bands) > 1:
        img_data_arr = numpy.moveaxis(img_data_arr, 0, -1)

    return img_data_arr, coords_bbox


def get_gdal_thematic_raster_mpl_imshow(
    input_img: str,
    band: int = 1,
    bbox: List[float] = None,
    out_patches=False,
    cls_names_lut=None,
) -> Tuple[numpy.array, List[float], list]:
    """
    A function which retrieves thematic image data with a colour table as an
    array in an appropriate structure for use within the matplotlib imshow function.
    The image pixel values are converted from there thematic integer values
    to a three band array using the RGB values from the colour table. If the
    pixel values are required then use the get_gdal_raster_mpl_imshow function.
    The extent is also returned and optionally a list of matplotlib patches
    which can be used to create a legend.

    :param input_img: The input image file path.
    :param band: The image band to be used for the visualisation (Default = 1).
    :param bbox: Optional bbox (xmin, xmax, ymin, ymax) used to subset the
                 input image so only data for the subset are returned.
    :param out_patches: Boolean to specify whether patches should be returned to
                        create a legend.
    :param cls_names_lut: A dictionary LUT with labels for the classes. The dict
                          key is the pixel value for the class and
    :return: numpy.array either [n,m,3], a bbox (xmin, xmax, ymin, ymax)
             specifying the extent of the image data and list of matplotlib patches,
             if out_patches=False then None is returned.

    .. code:: python

        img_sub_bbox = [554756, 577168, 9903924, 9944315]
        input_img = "class_img.kea"

        cls_names_lut = dict()
        cls_names_lut[1] = "Vegetation"
        cls_names_lut[2] = "Non-Veg"
        cls_names_lut[3] = "Productive Veg"

        (img_data_arr,
        coords_bbox,
        lgd_patches) = get_gdal_thematic_raster_mpl_imshow(input_img,
                                                           band=1,
                                                           bbox=img_sub_bbox,
                                                           out_patches=True,
                                                           cls_names_lut=cls_names_lut)


        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.imshow(img_data_arr, extent=coords_bbox)
        ax.legend(handles=lgd_patches)
        plt.show()

    """
    import rsgislib.imageutils

    n_bands = rsgislib.imageutils.get_img_band_count(input_img)
    if (band <= 0) or (band > n_bands):
        raise rsgislib.RSGISPyException(
            f"Band {band} is not valid (i.e., within the image)"
        )

    x_size, y_size = rsgislib.imageutils.get_img_size(input_img)
    pxl_bbox = [0, x_size, 0, y_size]
    if bbox is not None:
        pxl_bbox = rsgislib.imageutils.get_img_subset_pxl_bbox(input_img, bbox)

    n_x_pxls = pxl_bbox[1] - pxl_bbox[0]
    n_y_pxls = pxl_bbox[3] - pxl_bbox[2]

    print(f"Image Data Size: {n_x_pxls} x {n_y_pxls}")

    coords_bbox = rsgislib.imageutils.get_img_pxl_spatial_coords(input_img, pxl_bbox)

    image_ds = gdal.Open(input_img, gdal.GA_ReadOnly)
    if image_ds is None:
        raise rsgislib.RSGISPyException(
            f"Could not open the input image file: '{input_img}'"
        )

    clr_tab = image_ds.GetRasterBand(band).GetRasterColorTable()
    if clr_tab is None:
        raise rsgislib.RSGISPyException("No colour table was present")

    img_data_arr = image_ds.ReadAsArray(
        xoff=pxl_bbox[0],
        yoff=pxl_bbox[2],
        xsize=n_x_pxls,
        ysize=n_y_pxls,
        band_list=[band],
    )

    red_arr = numpy.zeros_like(img_data_arr, dtype=numpy.uint8)
    grn_arr = numpy.zeros_like(img_data_arr, dtype=numpy.uint8)
    blu_arr = numpy.zeros_like(img_data_arr, dtype=numpy.uint8)

    lgd_out_patches = None
    if out_patches:
        lgd_out_patches = list()

    for i in range(clr_tab.GetCount()):
        clr_tab_entry = clr_tab.GetColorEntry(i)
        red_arr[img_data_arr == i] = clr_tab_entry[0]
        grn_arr[img_data_arr == i] = clr_tab_entry[1]
        blu_arr[img_data_arr == i] = clr_tab_entry[2]

        if out_patches and (i > 0):
            cls_name = f"{i}"
            if (cls_names_lut is not None) and (i in cls_names_lut):
                cls_name = f"{cls_names_lut[i]}"
            rgb_clr = (
                clr_tab_entry[0] / 255.0,
                clr_tab_entry[1] / 255.0,
                clr_tab_entry[2] / 255.0,
            )
            lgd_out_patches.append(
                Patch(facecolor=rgb_clr, edgecolor=rgb_clr, label=cls_name)
            )

    img_clr_data_arr = numpy.stack([red_arr, grn_arr, blu_arr], axis=-1)

    image_ds = None
    img_data_arr = None

    return img_clr_data_arr, coords_bbox, lgd_out_patches


def limit_range_np_arr(
    arr_data: numpy.array,
    min_thres: float = 0,
    min_out_val: float = 0,
    max_thres: float = 1,
    max_out_val: float = 1,
) -> numpy.array:
    """
    A function which can be used to limit the range of the numpy array.
    For example, to mask values less than 0 to 0 and values greater than
    1 to 1.

    :param arr_data: input numpy array.
    :param min_thres: the threshold for the minimum value.
    :param min_out_val: the value assigned to values below the min_thres
    :param max_thres: the threshold for the maximum value.
    :param max_out_val: the value assigned to the values above the max_thres
    :return: numpy array with output values.

    """
    arr_data_out = arr_data.copy()
    arr_data_out[arr_data < min_thres] = min_out_val
    arr_data_out[arr_data > max_thres] = max_out_val
    return arr_data_out


def linear_stretch_np_arr(
    arr_data: numpy.array,
    no_data_val: float = None,
    out_off: float = 0,
    out_gain: float = 1,
    out_int_type=False,
    min_out_val: float = 0,
    max_out_val: float = 1,
) -> numpy.array:
    """
    A function which performs a linear stretch using the min-max values on a per
    band basis for a numpy array representing an image dataset. This function
    is useful in combination with get_gdal_raster_mpl_imshow for displaying
    raster data from an input image as a plot. By default this function returns
    values in a range 0 - 1 but if you prefer 0 - 255 then set the out_gain to
    255 and the out_int_type to be True to get an 8bit unsigned integer value.

    :param arr_data: The numpy array as either [n,m,b] or [n,m] where n and m are
                     the number of image pixels in the x and y axis' and b is the
                     number of image bands.
    :param no_data_val: the no data value for the input data. If there isn't a no
                        data value then leave as None (default)
    :param out_off: Output offset value (value * gain) + offset. Default: 0
    :param out_gain: Output gain value (value * gain) + offset. Default: 1
    :param out_int_type: False (default) and the output type will be float and
                         True and the output type with be integers.
    :param min_out_val: Minimum output value within the output array (default: 0)
    :param max_out_val: Maximum output value within the output array (default: 1)
    :return: A number array with the rescaled values but same dimensions as the
             input numpy array.

    .. code:: python

        img_sub_bbox = [554756, 577168, 9903924, 9944315]
        input_img = "sen2_img_strch.kea"

        img_data_arr, coords_bbox = get_gdal_raster_mpl_imshow(input_img,
                                                               bands=[8,9,3],
                                                               bbox=img_sub_bbox)

        img_data_arr = linear_stretch_np_arr(img_data_arr, no_data_val=0.0)


        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        im = ax.imshow(img_data_arr, extent=coords_bbox)
        plt.show()

    """
    arr_shp = arr_data.shape

    if no_data_val is not None:
        arr_data_out = arr_data.astype(float)
        arr_data_out[arr_data == no_data_val] = numpy.nan
    else:
        arr_data_out = arr_data.copy()

    if len(arr_shp) == 2:
        min_val = numpy.nanmin(arr_data_out)
        max_val = numpy.nanmax(arr_data_out)
        range_val = max_val - min_val

        arr_data_out = (((arr_data_out - min_val) / range_val) * out_gain) + out_off
    else:
        n_bands = arr_shp[2]
        for n in range(n_bands):
            min_val = numpy.nanmin(arr_data_out[..., n])
            max_val = numpy.nanmax(arr_data_out[..., n])
            range_val = max_val - min_val

            arr_data_out[..., n] = (
                ((arr_data_out[..., n] - min_val) / range_val) * out_gain
            ) + out_off

    arr_data_out = limit_range_np_arr(
        arr_data_out,
        min_thres=min_out_val,
        min_out_val=min_out_val,
        max_thres=max_out_val,
        max_out_val=max_out_val,
    )

    if out_int_type:
        arr_data_out = arr_data_out.astype(int)

    return arr_data_out


def cumulative_stretch_np_arr(
    arr_data: numpy.array,
    no_data_val: float = None,
    lower: int = 2,
    upper: int = 98,
    out_off: float = 0,
    out_gain: float = 1,
    out_int_type=False,
    min_out_val: float = 0,
    max_out_val: float = 1,
) -> numpy.array:
    """
    A function which performs a cumulative stretch using an upper and lower
    percentile to define the min-max values. This analysis is on a per
    band basis for a numpy array representing an image dataset. This function
    is useful in combination with get_gdal_raster_mpl_imshow for displaying
    raster data from an input image as a plot. By default this function returns
    values in a range 0 - 1 but if you prefer 0 - 255 then set the out_gain to
    255 and the out_int_type to be True to get an 8bit unsigned integer value.

    :param arr_data: The numpy array as either [n,m,b] or [n,m] where n and m are
                     the number of image pixels in the x and y axis' and b is the
                     number of image bands.
    :param no_data_val: the no data value for the input data. If there isn't a no
                        data value then leave as None (default)
    :param lower: lower percentile (default: 2)
    :param upper: upper percentile (default: 98)
    :param out_off: Output offset value (value * gain) + offset. Default: 0
    :param out_gain: Output gain value (value * gain) + offset. Default: 1
    :param out_int_type: False (default) and the output type will be float and
                         True and the output type with be integers.
    :param min_out_val: Minimum output value within the output array (default: 0)
    :param max_out_val: Maximum output value within the output array (default: 1)
    :return: A number array with the rescaled values but same dimensions as the
             input numpy array.

    .. code:: python

        img_sub_bbox = [554756, 577168, 9903924, 9944315]
        input_img = "sen2_img_strch.kea"

        img_data_arr, coords_bbox = get_gdal_raster_mpl_imshow(input_img,
                                                               bands=[8,9,3],
                                                               bbox=img_sub_bbox)

        img_data_arr = cumulative_stretch_np_arr(img_data_arr, no_data_val=0.0)

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        im = ax.imshow(img_data_arr, extent=coords_bbox)
        plt.show()

    """
    arr_shp = arr_data.shape

    if no_data_val is not None:
        arr_data_out = arr_data.astype(float)
        arr_data_out[arr_data == no_data_val] = numpy.nan
    else:
        arr_data_out = arr_data.copy()

    if len(arr_shp) == 2:
        min_val, max_val = numpy.nanpercentile(arr_data_out, [lower, upper])
        range_val = max_val - min_val

        arr_data_out = (((arr_data_out - min_val) / range_val) * out_gain) + out_off
    else:
        n_bands = arr_shp[2]
        for n in range(n_bands):
            min_val, max_val = numpy.nanpercentile(arr_data_out[..., n], [lower, upper])
            range_val = max_val - min_val

            arr_data_out[..., n] = (
                ((arr_data_out[..., n] - min_val) / range_val) * out_gain
            ) + out_off

    arr_data_out = limit_range_np_arr(
        arr_data_out,
        min_thres=min_out_val,
        min_out_val=min_out_val,
        max_thres=max_out_val,
        max_out_val=max_out_val,
    )

    if out_int_type:
        arr_data_out = arr_data_out.astype(int)

    return arr_data_out


def stdev_stretch_np_arr(
    arr_data: numpy.array,
    no_data_val: float = None,
    n_stdevs: float = 2.0,
    out_off: float = 0,
    out_gain: float = 1,
    out_int_type=False,
    min_out_val: float = 0,
    max_out_val: float = 1,
) -> numpy.array:
    """
    A function which performs a standard deviation stretch using an upper and lower
    (mean + n*std) and (mean - n*std) to define the min-max values. This analysis
    is on a per band basis for a numpy array representing an image dataset.
    This function is useful in combination with get_gdal_raster_mpl_imshow for
    displaying raster data from an input image as a plot. By default this function
    returns values in a range 0 - 1 but if you prefer 0 - 255 then set the out_gain
    to 255 and the out_int_type to be True to get an 8bit unsigned integer value.

    :param arr_data: The numpy array as either [n,m,b] or [n,m] where n and m are
                     the number of image pixels in the x and y axis' and b is the
                     number of image bands.
    :param no_data_val: the no data value for the input data. If there isn't a no
                        data value then leave as None (default)
    :param n_stdevs: number of standard deviations to be used for the stretch.
                     Default: 2.0
    :param out_off: Output offset value (value * gain) + offset. Default: 0
    :param out_gain: Output gain value (value * gain) + offset. Default: 1
    :param out_int_type: False (default) and the output type will be float and
                         True and the output type with be integers.
    :param min_out_val: Minimum output value within the output array (default: 0)
    :param max_out_val: Maximum output value within the output array (default: 1)
    :return: A number array with the rescaled values but same dimensions as the
             input numpy array.

    .. code:: python

        img_sub_bbox = [554756, 577168, 9903924, 9944315]
        input_img = "sen2_img_strch.kea"

        img_data_arr, coords_bbox = get_gdal_raster_mpl_imshow(input_img,
                                                               bands=[8,9,3],
                                                               bbox=img_sub_bbox)

        img_data_arr = stdev_stretch_np_arr(img_data_arr, no_data_val=0.0)

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        im = ax.imshow(img_data_arr, extent=coords_bbox)
        plt.show()

    """
    arr_shp = arr_data.shape

    if no_data_val is not None:
        arr_data_out = arr_data.astype(float)
        arr_data_out[arr_data == no_data_val] = numpy.nan
    else:
        arr_data_out = arr_data.copy()

    if len(arr_shp) == 2:
        min_val = numpy.nanmin(arr_data_out)
        max_val = numpy.nanmax(arr_data_out)
        mean_val = numpy.nanmean(arr_data_out)
        std_val = numpy.nanstd(arr_data_out)
        low_val = mean_val - (std_val * n_stdevs)
        up_val = mean_val + (std_val * n_stdevs)
        if low_val < min_val:
            low_val = min_val
        if up_val > max_val:
            up_val = max_val

        range_val = up_val - low_val

        arr_data_out = (((arr_data_out - low_val) / range_val) * out_gain) + out_off
    else:
        n_bands = arr_shp[2]
        for n in range(n_bands):
            min_val = numpy.nanmin(arr_data_out[..., n])
            max_val = numpy.nanmax(arr_data_out[..., n])
            mean_val = numpy.nanmean(arr_data_out[..., n])
            std_val = numpy.nanstd(arr_data_out[..., n])
            low_val = mean_val - (std_val * n_stdevs)
            up_val = mean_val + (std_val * n_stdevs)
            if low_val < min_val:
                low_val = min_val
            if up_val > max_val:
                up_val = max_val

            range_val = up_val - low_val

            arr_data_out[..., n] = (
                ((arr_data_out[..., n] - low_val) / range_val) * out_gain
            ) + out_off

    arr_data_out[arr_data_out < out_off] = out_off
    arr_data_out[arr_data_out > (out_gain + out_off)] = out_gain + out_off

    arr_data_out = limit_range_np_arr(
        arr_data_out,
        min_thres=min_out_val,
        min_out_val=min_out_val,
        max_thres=max_out_val,
        max_out_val=max_out_val,
    )

    if out_int_type:
        arr_data_out = arr_data_out.astype(int)

    return arr_data_out


def manual_stretch_np_arr(
    arr_data: numpy.array,
    min_max_vals: Union[Dict, List[Dict]],
    no_data_val: float = None,
    out_off: float = 0,
    out_gain: float = 1,
    out_int_type=False,
    min_out_val: float = 0,
    max_out_val: float = 1,
) -> numpy.array:
    """
    A function which performs a linear stretch using the min-max values provided
    on a per band basis for a numpy array representing an image dataset. This
    function is useful in combination with get_gdal_raster_mpl_imshow for
    displaying raster data from an input image as a plot. By default this function
    returns values in a range 0 - 1 but if you prefer 0 - 255 then set the out_gain
    to 255 and the out_int_type to be True to get an 8bit unsigned integer value.

    :param arr_data: The numpy array as either [n,m,b] or [n,m] where n and m are
                     the number of image pixels in the x and y axis' and b is the
                     number of image bands.
    :param min_max_vals: either a list of dicts each with a 'min' and 'max' key
                         specifying the min and max value for the stretch of each
                         band. Or, if just a single band then provide a single
                         dict rather than a list. The number items in the list
                         must equal the number of dimensions within the arr_data.
    :param no_data_val: the no data value for the input data. If there isn't a no
                        data value then leave as None (default)
    :param out_off: Output offset value (value * gain) + offset. Default: 0
    :param out_gain: Output gain value (value * gain) + offset. Default: 1
    :param out_int_type: False (default) and the output type will be float and
                         True and the output type with be integers.
    :param min_out_val: Minimum output value within the output array (default: 0)
    :param max_out_val: Maximum output value within the output array (default: 1)
    :return: A number array with the rescaled values but same dimensions as the
             input numpy array.

    .. code:: python

        img_sub_bbox = [554756, 577168, 9903924, 9944315]
        input_img = "sen2_img_strch.kea"

        img_data_arr, coords_bbox = get_gdal_raster_mpl_imshow(input_img,
                                                               bands=[8,9,3],
                                                               bbox=img_sub_bbox)

        min_max_vals = list()
        min_max_vals.append({'min':10, 'max':400})
        min_max_vals.append({'min':22, 'max':300})
        min_max_vals.append({'min':1, 'max':120})

        img_data_arr = manual_stretch_np_arr(img_data_arr,
                                             min_max_vals,
                                             no_data_val=0.0)


        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        im = ax.imshow(img_data_arr, extent=coords_bbox)
        plt.show()

    """
    arr_shp = arr_data.shape

    if no_data_val is not None:
        arr_data_out = arr_data.astype(float)
        arr_data_out[arr_data == no_data_val] = numpy.nan
    else:
        arr_data_out = arr_data.copy()

    if len(arr_shp) == 2:
        if type(min_max_vals) is not dict:
            raise rsgislib.RSGISPyException(
                "Just 1 dimension within arr_data and therefore "
                "min_max_vals variable must be a dict."
            )

        if ("min" not in min_max_vals) or ("max" not in min_max_vals):
            raise rsgislib.RSGISPyException(
                "min and max keys must be provided within the dict"
            )

        min_val = min_max_vals["min"]
        max_val = min_max_vals["max"]
        range_val = max_val - min_val

        arr_data_out = (((arr_data_out - min_val) / range_val) * out_gain) + out_off
    else:
        if type(min_max_vals) is not list:
            raise rsgislib.RSGISPyException(
                "arr_data has more than 1 dimension and therefore "
                "min_max_vals variable must be a list."
            )

        n_bands = arr_shp[2]

        if n_bands != len(min_max_vals):
            raise rsgislib.RSGISPyException(
                "length of min_max_vals must be the same as the number "
                "of bands in arr_data."
            )

        for n in range(n_bands):
            if ("min" not in min_max_vals[n]) or ("max" not in min_max_vals[n]):
                raise rsgislib.RSGISPyException(
                    "min and max keys must be provided within the dict"
                )

            min_val = min_max_vals[n]["min"]
            max_val = min_max_vals[n]["max"]
            range_val = max_val - min_val

            arr_data_out[..., n] = (
                ((arr_data_out[..., n] - min_val) / range_val) * out_gain
            ) + out_off

    arr_data_out = limit_range_np_arr(
        arr_data_out,
        min_thres=min_out_val,
        min_out_val=min_out_val,
        max_thres=max_out_val,
        max_out_val=max_out_val,
    )

    if out_int_type:
        arr_data_out = arr_data_out.astype(int)

    return arr_data_out
